# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import uuid


LOG = logging.getLogger(__name__)


def _process_params(flattened, params):
    for item in params:
        if item not in flattened['parameters']:
            param_obj = {}
            for key, value in params.get(item).items():
                camel_case_key = key[0].lower() + key[1:]
                param_obj[camel_case_key] = value
            param_obj['name'] = item
            flattened['parameters'][item] = param_obj
    return list(params)


def _flat_it(flattened, name, data):
    key = str(uuid.uuid4())
    value = {}
    value.update({
        'name': name,
        'id': key
    })
    if 'Type' in data:
        value['type'] = data['Type']
    if 'Description' in data:
        value['description'] = data['Description']
    if 'Parameters' in data:
        value['parameters'] = _process_params(flattened,
                                              data['Parameters'])
    if 'ParameterGroups' in data:
        value['parameter_groups'] = data['ParameterGroups']
    if 'NestedParameters' in data:
        nested = data['NestedParameters']
        nested_ids = []
        for nested_key in nested.keys():
            nested_data = _flat_it(flattened, nested_key,
                                   nested.get(nested_key))
            # nested_data will always have one key (and only one)
            nested_ids.append(list(nested_data)[0])

        value['resources'] = nested_ids

    flattened['resources'][key] = value
    return {key: value}


def preview_stack_and_network_configs(heat, processed_data,
                                      container, role_name):
    # stacks.preview method raises validation message if stack is
    # already deployed. here renaming container to get preview data.
    container_temp = container + "-TEMP"
    fields = {
        'template': processed_data['template'],
        'files': processed_data['files'],
        'environment': processed_data['environment'],
        'stack_name': container_temp,
    }
    preview_data = heat.stacks.preview(**fields)
    return get_network_config(preview_data, container_temp, role_name)


def get_network_config(preview_data, stack_name, role_name):
    result = None
    if preview_data:
        for res in preview_data.resources:
            net_script = process_preview_list(res, stack_name,
                                              role_name)
            if net_script:
                ns_len = len(net_script)
                start_index = (net_script.find(
                    "echo '{\"network_config\"", 0, ns_len) + 6)
                # In file network/scripts/run-os-net-config.sh
                end_str = "' > /etc/os-net-config/config.json"
                end_index = net_script.find(end_str, start_index, ns_len)
                if (end_index > start_index):
                    net_config = net_script[start_index:end_index]
                    if net_config:
                        result = json.loads(net_config)
                break
    if not result:
        err_msg = ("Unable to determine network config for role '%s'."
                   % role_name)
        LOG.exception(err_msg)
        raise RuntimeError(err_msg)
    return result


def process_preview_list(res, stack_name, role_name):
    if type(res) == list:
        for item in res:
            out = process_preview_list(item, stack_name, role_name)
            if out:
                return out
    elif type(res) == dict:
        res_stack_name = stack_name + '-' + role_name
        if res['resource_name'] == "OsNetConfigImpl" and \
            res['resource_identity'] and \
            res_stack_name in res['resource_identity']['stack_name']:
            return res['properties']['config']
    return None



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import OrderedDict
import fixtures
import os
import sys
from unittest import mock

import yaml

from heatclient.exc import HTTPNotFound

from tripleo_common.inventory import NeutronData
from tripleo_common.inventory import StackOutputs
from tripleo_common.inventory import TripleoInventory
from tripleo_common.tests import base
from tripleo_common.tests.fake_neutron import fakes as neutron_fakes


MOCK_ENABLED_SERVICES = {
    "ObjectStorage": [
        "kernel",
        "swift_storage",
        "tripleo_packages"
        ],
    "Controller": [
        "kernel",
        "keystone",
        "tripleo_packages"
        ],
    "Compute": [
        "nova_compute",
        "kernel",
        "tripleo_packages",
        "ceph_client"
        ],
    "CephStorage": [
        "kernel",
        "tripleo_packages"
        ],
    "BlockStorage": [
        "cinder_volume",
        "kernel",
        "tripleo_packages"
        ]
    }


class TestInventory(base.TestCase):
    def setUp(self):
        super(TestInventory, self).setUp()
        self.outputs_data = {'outputs': [
            {
                'output_key': 'EnabledServices',
                'output_value': {
                    'Controller': ['sa', 'sb'],
                    'Compute': ['sd', 'se', 'ceph_client'],
                    'CustomRole': ['sg', 'sh']}
                },
            {
                'output_key': 'KeystoneURL',
                'output_value': 'xyz://keystone'
                },
            {
                'output_key': 'ServerIdData',
                'output_value': {
                    'server_ids': {
                        'Controller': ['a', 'b', 'c'],
                        'Compute': ['d'],
                        'CustomRole': ['e']
                        },
                    'bootstrap_server_id': 'a'
                    }
                },
            {
                'output_key': 'RoleNetHostnameMap',
                'output_value': {
                    'Controller': {
                        'ctlplane': [
                            'c-0.ctlplane.localdomain',
                            'c-1.ctlplane.localdomain',
                            'c-2.ctlplane.localdomain'],
                        'internal_api': [
                            'c-0.internal_api.localdomain',
                            'c-1.internal_api.localdomain',
                            'c-2.internal_api.localdomain']
                        },
                    'Compute': {
                        'ctlplane': ['cp-0.ctlplane.localdomain']
                        },
                    'CustomRole': {
                        'ctlplane': ['cs-0.ctlplane.localdomain']
                        }
                    }
                },
            {
                'output_key': 'RoleNetIpMap',
                'output_value': {
                    'Controller': {
                        'ctlplane': [
                            'x.x.x.1',
                            'x.x.x.2',
                            'x.x.x.3'
                            ],
                        'internal_api': [
                            'x.x.x.4',
                            'x.x.x.5',
                            'x.x.x.6'
                            ]
                        },
                    'Compute': {
                        'ctlplane': ['y.y.y.1']
                        },
                    'CustomRole': {
                        'ctlplane': ['z.z.z.1']
                        }
                    }
                },
            {
                'output_key': 'VipMap',
                'output_value': {
                    'ctlplane': 'x.x.x.4',
                    'redis': 'x.x.x.6'
                    }
                },
            {
                'output_key': 'RoleData',
                'output_value': {
                    'Controller': {'config_settings': 'foo1'},
                    'Compute': {'config_settings': 'foo2'},
                    'CustomRole': {'config_settings': 'foo3'}
                    }
                }
            ]
        }
        self.plan_name = 'overcloud'

        self.hclient = mock.MagicMock()
        self.hclient.stacks.environment.return_value = {
            'parameter_defaults': {
                'AdminPassword': 'theadminpw',
                'ContainerCli': 'podman'
                }
            }
        self.mock_stack = mock.MagicMock()
        self.mock_stack.outputs = self.outputs_data['outputs']
        self.hclient.stacks.get.return_value = self.mock_stack

        self.outputs = StackOutputs(self.mock_stack)
        self.inventory = TripleoInventory(
            hclient=self.hclient,
            plan_name=self.plan_name,
            auth_url='xyz://keystone.local',
            cacert='acacert',
            project_name='admin',
            username='admin',
            ansible_ssh_user='heat-admin')
        self.inventory.stack_outputs = self.outputs

    def test_get_roles_by_service(self):
        services = TripleoInventory.get_roles_by_service(MOCK_ENABLED_SERVICES)
        expected = {
            'kernel': [
                'BlockStorage',
                'CephStorage',
                'Compute',
                'Controller',
                'ObjectStorage'
                ],
            'swift_storage': ['ObjectStorage'],
            'tripleo_packages': [
                'BlockStorage',
                'CephStorage',
                'Compute',
                'Controller',
                'ObjectStorage'
                ],
            'keystone': ['Controller'],
            'nova_compute': ['Compute'],
            'cinder_volume': ['BlockStorage'],
            'ceph_client': ['Compute'],
        }
        self.assertDictEqual(services, expected)

    def test_stack_not_found(self):
        self.hclient.stacks.get.side_effect = HTTPNotFound('not found')
        self.assertEqual(None, self.inventory._get_stack())

    def test_outputs_valid_key_calls_api(self):
        expected = 'xyz://keystone'
        self.hclient.stacks.output_show.return_value = dict(output=dict(
            output_value=expected))
        self.assertEqual(expected, self.outputs['KeystoneURL'])
        # This should also support the get method
        self.assertEqual(expected, self.outputs.get('KeystoneURL'))
        self.assertTrue(self.hclient.called_once_with('overcloud',
                                                      'KeystoneURL'))

    def test_no_ips(self):
        for output in self.outputs_data['outputs']:
            if output['output_key'] == 'RoleNetIpMap':
                output['output_value'] = dict(Controller=dict(ctlplane=[]))
        self.assertRaises(Exception, self.inventory.list)

    def test_outputs_invalid_key_raises_keyerror(self):
        self.assertRaises(KeyError, lambda: self.outputs['Invalid'])

    def test_outputs_get_method_returns_default(self):
        default = 'default value'
        self.assertEqual(default, self.outputs.get('Invalid', default))

    def test_outputs_iterating_returns_list_of_output_keys(self):
        self.assertEqual({
            'EnabledServices',
            'KeystoneURL',
            'ServerIdData',
            'RoleNetHostnameMap',
            'RoleNetIpMap',
            'VipMap',
            'RoleData'
        }, set([o for o in self.outputs]))

    def test_inventory_list(self):
        self.inventory.undercloud_connection = 'local'
        self._inventory_list(self.inventory)

    def _inventory_list(self, inventory):
        ansible_ssh_user = 'heat-admin'
        expected = {
            'Compute': {
                'hosts': ['cp-0'],
                'vars': {
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'Compute',
                    'tripleo_role_networks': ['ctlplane']
                    }
                },
            'Controller': {
                'hosts': ['c-0', 'c-1', 'c-2'],
                'vars': {
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'Controller',
                    'tripleo_role_networks': [
                        'ctlplane',
                        'internal_api'
                        ]
                    }
                },
            'CustomRole': {
                'hosts': ['cs-0'],
                'vars': {
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'CustomRole',
                    'tripleo_role_networks': ['ctlplane']
                    }
                },
            'overcloud': {
                'children': ['allovercloud']
                },
            'allovercloud': {
                'children': ['Compute', 'Controller', 'CustomRole'],
                'vars': {
                    'container_cli': 'podman',
                    'ctlplane_vip': 'x.x.x.4',
                    'redis_vip': 'x.x.x.6'
                    }
                },
            'Undercloud': {
                'hosts': ['undercloud'],
                'vars': {
                    'ansible_connection': 'local',
                    'ansible_host': 'localhost',
                    'ansible_python_interpreter': sys.executable,
                    'ansible_remote_tmp': '/tmp/ansible-${USER}',
                    'auth_url': 'xyz://keystone.local',
                    'cacert': 'acacert',
                    'overcloud_keystone_url': 'xyz://keystone',
                    'overcloud_admin_password': 'theadminpw',
                    'plan': 'overcloud',
                    'project_name': 'admin',
                    'undercloud_service_list': [
                        'tripleo_nova_compute',
                        'tripleo_heat_engine',
                        'tripleo_ironic_conductor',
                        'tripleo_swift_container_server',
                        'tripleo_swift_object_server',
                        'tripleo_mistral_engine'
                        ],
                    'username': 'admin'
                    }
                }
            }
        inv_list = inventory.list()
        for k in expected:
            self.assertEqual(expected[k], inv_list[k])

    def test_inventory_list_undercloud_installer(self):
        outputs_data = {
            'outputs': [
                {'output_key': 'EnabledServices',
                 'output_value': {'Undercloud': ['sa', 'sb']}},
                {'output_key': 'KeystoneURL',
                 'output_value': 'xyz://keystone'},
                {'output_key': 'ServerIdData',
                 'output_value': {'server_ids': {'Undercloud': ['a']},
                                  'bootstrap_server_id': 'a'}},
                {'output_key': 'RoleNetHostnameMap',
                 'output_value': {'Undercloud': {
                     'ctlplane': ['uc0.ctlplane.localdomain'],
                     'external': ['uc0.external.localdomain'],
                     'canonical': ['uc0.lab.example.com']}}},
                {'output_key': 'RoleNetIpMap',
                 'output_value': {'Undercloud': {'ctlplane': ['x.x.x.1'],
                                                 'external': ['x.x.x.1']}}},
                {'output_key': 'VipMap',
                 'output_value': {'ctlplane': 'x.x.x.4', 'redis': 'x.x.x.6'}},
                {'output_key': 'RoleData',
                 'output_value': {'Undercloud': {'config_settings': 'foo1'}}}
            ]
        }
        plan_name = 'undercloud'
        hclient = mock.MagicMock()
        hclient.stacks.environment.return_value = {'parameter_defaults': {
            'AdminPassword': 'theadminpw', 'ContainerCli': 'podman'}}
        mock_stack = mock.MagicMock()
        mock_stack.outputs = outputs_data['outputs']
        hclient.stacks.get.return_value = mock_stack

        outputs = StackOutputs(mock_stack)
        inventory = TripleoInventory(
            hclient=hclient,
            plan_name=plan_name,
            auth_url='xyz://keystone.local',
            cacert='acacert',
            project_name='admin',
            username='admin',
            ansible_ssh_user='heat-admin')
        inventory.stack_outputs = outputs
        expected = {
            'Undercloud': {
                'hosts': {
                    'uc0': {
                        'ansible_host': 'x.x.x.1',
                        'canonical_hostname': 'uc0.lab.example.com',
                        'ctlplane_hostname': 'uc0.ctlplane.localdomain',
                        'ctlplane_ip': 'x.x.x.1',
                        'deploy_server_id': 'a',
                        'external_hostname': 'uc0.external.localdomain',
                        'external_ip': 'x.x.x.1'}},
                'vars': {
                    'ansible_ssh_user': 'heat-admin',
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'Undercloud',
                    'tripleo_role_networks': ['ctlplane', 'external']}},
            'allovercloud': {
                'children': {'Undercloud': {}},
                'vars': {'container_cli': 'podman',
                         'ctlplane_vip': 'x.x.x.4',
                         'redis_vip': 'x.x.x.6'}},
            'sb': {'children': {'Undercloud': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}},
            'sa': {'children': {'Undercloud': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}}
            }
        inv_list = inventory.list(dynamic=False)
        for k in expected:
            self.assertEqual(expected[k], inv_list[k])

    def test_inventory_list_undercloud_only(self):
        self.inventory.plan_name = None
        self.inventory.undercloud_connection = 'local'
        expected = {
            'Undercloud': {
                'hosts': ['undercloud'],
                'vars': {
                    'ansible_connection': 'local',
                    'ansible_host': 'localhost',
                    'ansible_python_interpreter': sys.executable,
                    'ansible_remote_tmp': '/tmp/ansible-${USER}',
                    'auth_url': 'xyz://keystone.local',
                    'cacert': 'acacert',
                    'project_name': 'admin',
                    'undercloud_service_list': [
                        'tripleo_nova_compute',
                        'tripleo_heat_engine',
                        'tripleo_ironic_conductor',
                        'tripleo_swift_container_server',
                        'tripleo_swift_object_server',
                        'tripleo_mistral_engine'
                        ],
                    'username': 'admin'
                }
            },
            '_meta': {'hostvars': {}},
        }
        self.assertEqual(expected, self.inventory.list())

    def test_ansible_ssh_user(self):
        self._try_alternative_args(
            ansible_ssh_user='my-custom-admin', undercloud_connection='ssh')

    def _try_alternative_args(self, ansible_ssh_user, undercloud_connection):
        key_file = '/var/lib/mistral/.ssh/%s-key' % ansible_ssh_user
        self.inventory = TripleoInventory(
            hclient=self.hclient,
            plan_name=self.plan_name,
            auth_url='xyz://keystone.local',
            project_name='admin',
            username='admin',
            cacert='acacert',
            ansible_ssh_user=ansible_ssh_user,
            undercloud_connection=undercloud_connection,
            undercloud_key_file=key_file,
            ansible_python_interpreter='foo'
        )

        self.inventory.stack_outputs = self.outputs

        expected = {
            'Compute': {
                'hosts': ['cp-0'],
                'vars': {
                    'ansible_python_interpreter': 'foo',
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'Compute',
                    'tripleo_role_networks': ['ctlplane']
                    }
                },
            'Controller': {
                'hosts': ['c-0', 'c-1', 'c-2'],
                'vars': {
                    'ansible_python_interpreter': 'foo',
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'Controller',
                    'tripleo_role_networks': [
                        'ctlplane',
                        'internal_api'
                        ]
                    }
                },
            'CustomRole': {
                'hosts': ['cs-0'],
                'vars': {
                    'ansible_python_interpreter': 'foo',
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'CustomRole',
                    'tripleo_role_networks': ['ctlplane']
                    }
            },
            'overcloud': {
                'children': ['allovercloud']
            },
            'allovercloud': {
                'children': ['Compute', 'Controller', 'CustomRole'],
                'vars': {
                    'container_cli': 'podman',
                    'ctlplane_vip': 'x.x.x.4',
                    'redis_vip': 'x.x.x.6'
                }
            },
            'sa': {
                'children': ['Controller'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                }
            },
            'sb': {
                'children': ['Controller'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                }
            },
            'sd': {
                'children': ['Compute'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                }
            },
            'se': {
                'children': ['Compute'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                }
            },
            'ceph_client': {
                'children': ['Compute'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                }
            },
            'clients': {
                'children': ['Compute'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                    }
                },
            'sg': {
                'children': ['CustomRole'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                }
            },
            'sh': {
                'children': ['CustomRole'],
                'vars': {
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_python_interpreter': 'foo'
                    }
                },
            'Undercloud': {
                'hosts': ['undercloud'],
                'vars': {
                    'ansible_connection': 'ssh',
                    'ansible_ssh_private_key_file': key_file,
                    'ansible_ssh_user': 'my-custom-admin',
                    'ansible_host': 'localhost',
                    'ansible_python_interpreter': 'foo',
                    'ansible_remote_tmp': '/tmp/ansible-${USER}',
                    'auth_url': 'xyz://keystone.local',
                    'cacert': 'acacert',
                    'overcloud_keystone_url': 'xyz://keystone',
                    'overcloud_admin_password': 'theadminpw',
                    'plan': 'overcloud',
                    'project_name': 'admin',
                    'undercloud_service_list': [
                        'tripleo_nova_compute',
                        'tripleo_heat_engine',
                        'tripleo_ironic_conductor',
                        'tripleo_swift_container_server',
                        'tripleo_swift_object_server',
                        'tripleo_mistral_engine'],
                    'username': 'admin'
                    }
                }
            }

        inv_list = self.inventory.list()
        for k in expected:
            self.assertEqual(expected[k], inv_list[k])

    def test_inventory_write_static(self):
        self.inventory.undercloud_connection = 'local'
        self._inventory_write_static()

    def test_inventory_write_static_extra_vars(self):
        self.inventory.undercloud_connection = 'local'
        extra_vars = {'Undercloud': {'anextravar': 123}}
        self._inventory_write_static(extra_vars=extra_vars)

    def _inventory_write_static(self, extra_vars=None):
        tmp_dir = self.useFixture(fixtures.TempDir()).path
        inv_path = os.path.join(tmp_dir, "inventory.yaml")
        self.inventory.write_static_inventory(inv_path, extra_vars)
        ansible_ssh_user = 'heat-admin'
        expected = {
            'Undercloud': {
                'hosts': {'undercloud': {}},
                'vars': {
                    'ansible_connection': 'local',
                    'ansible_host': 'localhost',
                    'ansible_python_interpreter':
                    sys.executable,
                    'ansible_remote_tmp':
                    '/tmp/ansible-${USER}',
                    'auth_url': 'xyz://keystone.local',
                    'cacert': 'acacert',
                    'overcloud_admin_password': 'theadminpw',
                    'overcloud_keystone_url': 'xyz://keystone',
                    'plan': 'overcloud',
                    'project_name': 'admin',
                    'undercloud_service_list': [
                        'tripleo_nova_compute',
                        'tripleo_heat_engine',
                        'tripleo_ironic_conductor',
                        'tripleo_swift_container_server',
                        'tripleo_swift_object_server',
                        'tripleo_mistral_engine'],
                    'username': 'admin'
                }
            },
            'Controller': {
                'hosts': {
                    'c-0': {
                        'ansible_host': 'x.x.x.1',
                        'ctlplane_ip': 'x.x.x.1',
                        'deploy_server_id': 'a',
                        'ctlplane_hostname': 'c-0.ctlplane.localdomain',
                        'internal_api_hostname':
                        'c-0.internal_api.localdomain',
                        'internal_api_ip': 'x.x.x.4'
                    },
                    'c-1': {
                        'ansible_host': 'x.x.x.2',
                        'ctlplane_ip': 'x.x.x.2',
                        'deploy_server_id': 'b',
                        'ctlplane_hostname': 'c-1.ctlplane.localdomain',
                        'internal_api_hostname':
                        'c-1.internal_api.localdomain',
                        'internal_api_ip': 'x.x.x.5'
                    },
                    'c-2': {
                        'ansible_host': 'x.x.x.3',
                        'ctlplane_ip': 'x.x.x.3',
                        'deploy_server_id': 'c',
                        'ctlplane_hostname': 'c-2.ctlplane.localdomain',
                        'internal_api_hostname':
                        'c-2.internal_api.localdomain',
                        'internal_api_ip': 'x.x.x.6'
                    }
                },
                'vars': {
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'Controller',
                    'tripleo_role_networks': [
                        'ctlplane',
                        'internal_api'
                    ]
                }
            },
            'Compute': {
                'hosts': {
                    'cp-0': {
                        'ansible_host': 'y.y.y.1',
                        'ctlplane_ip': 'y.y.y.1',
                        'deploy_server_id': 'd',
                        'ctlplane_hostname': 'cp-0.ctlplane.localdomain'
                    }
                },
                'vars': {
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'Compute',
                    'tripleo_role_networks': ['ctlplane']
                }
            },
            'CustomRole': {
                'hosts': {
                    'cs-0': {
                        'ansible_host': 'z.z.z.1',
                        'ctlplane_ip': 'z.z.z.1',
                        'deploy_server_id': 'e',
                        'ctlplane_hostname': 'cs-0.ctlplane.localdomain'
                    }
                },
                'vars': {
                    'ansible_ssh_user': ansible_ssh_user,
                    'bootstrap_server_id': 'a',
                    'serial': 1,
                    'tripleo_role_name': 'CustomRole',
                    'tripleo_role_networks': ['ctlplane']
                }
            },
            'overcloud': {
                'children': {'allovercloud': {}}
            },
            'allovercloud': {
                'children': {
                    'Compute': {},
                    'Controller': {},
                    'CustomRole': {}
                },
                'vars': {
                    'container_cli': 'podman',
                    'ctlplane_vip': 'x.x.x.4',
                    'redis_vip': 'x.x.x.6'
                }
            },
            'sa': {
                'children': {'Controller': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
            'sb': {
                'children': {'Controller': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
            'sd': {
                'children': {'Compute': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
            'se': {
                'children': {'Compute': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
            'ceph_client': {
                'children': {'Compute': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
            'clients': {
                'children': {'Compute': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
            'sg': {
                'children': {'CustomRole': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
            'sh': {
                'children': {'CustomRole': {}},
                'vars': {'ansible_ssh_user': 'heat-admin'}
            },
        }
        if extra_vars:
            expected['Undercloud']['vars']['anextravar'] = 123

        with open(inv_path, 'r') as f:
            loaded_inv = yaml.safe_load(f)
        self.assertEqual(expected, loaded_inv)

    def test__add_host_from_neutron_data(self):
        neutron_data = NeutronData(networks=neutron_fakes.fake_networks,
                                   subnets=neutron_fakes.fake_subnets,
                                   ports=neutron_fakes.compute_0_ports)
        ret = OrderedDict()
        role = ret.setdefault('Compute', {})
        role_vars = role.setdefault('vars', {})
        role_networks = role_vars.setdefault('tripleo_role_networks', [])
        hosts = role.setdefault('hosts', {})
        ports = neutron_data.ports_by_role_and_host['Compute']['cp-0']
        self.inventory._add_host_from_neutron_data(hosts, ports, role_networks,
                                                   role_vars)
        self.assertEqual(OrderedDict([
            ('Compute',
             {'hosts': {
                 'ansible_host': '192.0.2.20',
                 'canonical_hostname': 'cp-0.example.com.',
                 'ctlplane_hostname': 'cp-0.ctlplane.example.com.',
                 'ctlplane_ip': '192.0.2.20',
                 'internal_api_hostname': 'cp-0.internalapi.example.com',
                 'internal_api_ip': '198.51.100.150'},
                 'vars': {
                     'ctlplane_cidr': '24',
                     'ctlplane_dns_nameservers': ['192.0.2.253',
                                                  '192.0.2.254'],
                     'ctlplane_gateway_ip': '192.0.2.1',
                     'ctlplane_host_routes': [{'default': True,
                                               'nexthop': '192.0.2.1'}],
                     'ctlplane_vlan_id': '1',
                     'internal_api_cidr': '25',
                     'internal_api_dns_nameservers': [],
                     'internal_api_gateway_ip': '198.51.100.129',
                     'internal_api_host_routes': [],
                     'internal_api_vlan_id': '20',
                     'tripleo_role_networks': ['ctlplane', 'internal_api']
                 }})
        ]), ret)

    def test__inventory_from_neutron_data(self):
        ret = OrderedDict()
        children = set()
        fake_ports = (neutron_fakes.controller0_ports +
                      neutron_fakes.controller1_ports +
                      neutron_fakes.compute_0_ports)
        self.inventory.neutron_data = NeutronData(
            networks=neutron_fakes.fake_networks,
            subnets=neutron_fakes.fake_subnets,
            ports=fake_ports)

        self.inventory._inventory_from_neutron_data(ret, children, False)
        self.assertEqual({'Compute', 'Controller'}, children)
        self.assertEqual(OrderedDict([
            ('Controller',
             {'hosts': {
                 'c-0': {
                     'ansible_host': '192.0.2.10',
                     'canonical_hostname': 'c-0.example.com.',
                     'ctlplane_hostname': 'c-0.ctlplane.example.com.',
                     'ctlplane_ip': '192.0.2.10',
                     'internal_api_hostname': 'c-0.internalapi.example.com',
                     'internal_api_ip': '198.51.100.140'},
                 'c-1': {
                     'ansible_host': '192.0.2.11',
                     'canonical_hostname': 'c-1.example.com.',
                     'ctlplane_hostname': 'c-1.ctlplane.example.com.',
                     'ctlplane_ip': '192.0.2.11',
                     'internal_api_hostname': 'c-1.internalapi.example.com',
                     'internal_api_ip': '198.51.100.141'}},
                 'vars': {'ansible_ssh_user': 'heat-admin',
                          'ctlplane_cidr': '24',
                          'ctlplane_dns_nameservers': ['192.0.2.253',
                                                       '192.0.2.254'],
                          'ctlplane_gateway_ip': '192.0.2.1',
                          'ctlplane_host_routes': [{'default': True,
                                                    'nexthop': '192.0.2.1'}],
                          'ctlplane_mtu': 1500,
                          'ctlplane_subnet_cidr': '24',
                          'ctlplane_vlan_id': '1',
                          'internal_api_cidr': '25',
                          'internal_api_dns_nameservers': [],
                          'internal_api_gateway_ip': '198.51.100.129',
                          'internal_api_host_routes': [],
                          'internal_api_mtu': 1500,
                          'internal_api_vlan_id': '20',
                          'networks_all': ['InternalApi'],
                          'networks_lower': {'InternalApi': 'internal_api',
                                             'ctlplane': 'ctlplane'},
                          'role_networks': ['ctlplane', 'InternalApi'],
                          'serial': 1,
                          'tripleo_role_name': 'Controller',
                          'tripleo_role_networks': ['ctlplane', 'internal_api']
                          }}),
            ('Compute',
             {'hosts': {
                 'cp-0': {
                     'ansible_host': '192.0.2.20',
                     'canonical_hostname': 'cp-0.example.com.',
                     'ctlplane_hostname': 'cp-0.ctlplane.example.com.',
                     'ctlplane_ip': '192.0.2.20',
                     'internal_api_hostname': 'cp-0.internalapi.example.com',
                     'internal_api_ip': '198.51.100.150'}},
                 'vars': {'ansible_ssh_user': 'heat-admin',
                          'ctlplane_cidr': '24',
                          'ctlplane_dns_nameservers': ['192.0.2.253',
                                                       '192.0.2.254'],
                          'ctlplane_gateway_ip': '192.0.2.1',
                          'ctlplane_host_routes': [{'default': True,
                                                    'nexthop': '192.0.2.1'}],
                          'ctlplane_mtu': 1500,
                          'ctlplane_subnet_cidr': '24',
                          'ctlplane_vlan_id': '1',
                          'internal_api_cidr': '25',
                          'internal_api_dns_nameservers': [],
                          'internal_api_gateway_ip': '198.51.100.129',
                          'internal_api_host_routes': [],
                          'internal_api_mtu': 1500,
                          'internal_api_vlan_id': '20',
                          'networks_all': ['InternalApi'],
                          'networks_lower': {'InternalApi': 'internal_api',
                                             'ctlplane': 'ctlplane'},
                          'role_networks': ['ctlplane', 'InternalApi'],
                          'serial': 1,
                          'tripleo_role_name': 'Compute',
                          'tripleo_role_networks': ['ctlplane', 'internal_api']
                          }}),
            ('allovercloud', {'children': {'Compute': {}, 'Controller': {}}})
        ]), ret)

    def test__inventory_from_neutron_data_dynamic(self):
        ret = OrderedDict()
        children = set()
        fake_ports = (neutron_fakes.controller0_ports +
                      neutron_fakes.controller1_ports +
                      neutron_fakes.compute_0_ports)
        self.inventory.neutron_data = NeutronData(
            networks=neutron_fakes.fake_networks,
            subnets=neutron_fakes.fake_subnets,
            ports=fake_ports)

        self.inventory._inventory_from_neutron_data(ret, children, True)
        self.assertEqual({'Compute', 'Controller'}, children)
        self.assertEqual(OrderedDict([
            ('Controller', {
                'hosts': ['c-0', 'c-1'],
                'vars': {'ansible_ssh_user': 'heat-admin',
                         'ctlplane_cidr': '24',
                         'ctlplane_dns_nameservers': ['192.0.2.253',
                                                      '192.0.2.254'],
                         'ctlplane_gateway_ip': '192.0.2.1',
                         'ctlplane_host_routes': [{'default': True,
                                                   'nexthop': '192.0.2.1'}],
                         'ctlplane_mtu': 1500,
                         'ctlplane_vlan_id': '1',
                         'internal_api_cidr': '25',
                         'internal_api_dns_nameservers': [],
                         'internal_api_gateway_ip': '198.51.100.129',
                         'internal_api_host_routes': [],
                         'internal_api_mtu': 1500,
                         'ctlplane_subnet_cidr': '24',
                         'internal_api_vlan_id': '20',
                         'networks_all': ['InternalApi'],
                         'networks_lower': {'InternalApi': 'internal_api',
                                            'ctlplane': 'ctlplane'},
                         'role_networks': ['ctlplane', 'InternalApi'],
                         'serial': 1,
                         'tripleo_role_name': 'Controller',
                         'tripleo_role_networks': ['ctlplane', 'internal_api']
                         }}),
            ('Compute', {
                'hosts': ['cp-0'],
                'vars': {'ansible_ssh_user': 'heat-admin',
                         'ctlplane_cidr': '24',
                         'ctlplane_dns_nameservers': ['192.0.2.253',
                                                      '192.0.2.254'],
                         'ctlplane_gateway_ip': '192.0.2.1',
                         'ctlplane_host_routes': [{'default': True,
                                                   'nexthop': '192.0.2.1'}],
                         'ctlplane_mtu': 1500,
                         'ctlplane_vlan_id': '1',
                         'internal_api_cidr': '25',
                         'internal_api_dns_nameservers': [],
                         'internal_api_gateway_ip': '198.51.100.129',
                         'internal_api_host_routes': [],
                         'internal_api_mtu': 1500,
                         'ctlplane_subnet_cidr': '24',
                         'internal_api_vlan_id': '20',
                         'networks_all': ['InternalApi'],
                         'networks_lower': {'InternalApi': 'internal_api',
                                            'ctlplane': 'ctlplane'},
                         'role_networks': ['ctlplane', 'InternalApi'],
                         'serial': 1,
                         'tripleo_role_name': 'Compute',
                         'tripleo_role_networks': ['ctlplane', 'internal_api']
                         }}),
            ('allovercloud', {'children': ['Compute', 'Controller']})]
        ), ret)

    @mock.patch.object(TripleoInventory, '_get_neutron_data', autospec=True)
    def test_inventory_list_with_neutron_and_heat(self, mock_get_neutron_data):
        fake_ports = (neutron_fakes.controller0_ports +
                      neutron_fakes.controller1_ports +
                      neutron_fakes.controller2_ports +
                      neutron_fakes.compute_0_ports +
                      neutron_fakes.custom_0_ports)
        mock_get_neutron_data.return_value = NeutronData(
            networks=neutron_fakes.fake_networks,
            subnets=neutron_fakes.fake_subnets,
            ports=fake_ports)
        inv_list = self.inventory.list(dynamic=False)
        c_0 = inv_list['Controller']['hosts']['c-0']
        c_1 = inv_list['Controller']['hosts']['c-1']
        c_2 = inv_list['Controller']['hosts']['c-2']
        cp_0 = inv_list['Compute']['hosts']['cp-0']
        cs_0 = inv_list['CustomRole']['hosts']['cs-0']

        # The setdefault pattern should always put the value discovered first
        # in the inventory, neutron source run's prior to heat stack source.
        # Assert IP addresses from neutron fake are used in the
        # inventory, not the heat stack IPs.

        # Controller
        self.assertNotEqual(
            c_0['ctlplane_ip'],
            self.outputs['RoleNetIpMap']['Controller']['ctlplane'][0])
        self.assertNotEqual(
            c_0['ansible_host'],
            self.outputs['RoleNetIpMap']['Controller']['ctlplane'][0])
        self.assertNotEqual(
            c_1['ctlplane_ip'],
            self.outputs['RoleNetIpMap']['Controller']['ctlplane'][1])
        self.assertNotEqual(
            c_1['ansible_host'],
            self.outputs['RoleNetIpMap']['Controller']['ctlplane'][1])
        self.assertNotEqual(
            c_2['ctlplane_ip'],
            self.outputs['RoleNetIpMap']['Controller']['ctlplane'][2])
        self.assertNotEqual(
            c_2['ansible_host'],
            self.outputs['RoleNetIpMap']['Controller']['ctlplane'][2])
        # Compute
        self.assertNotEqual(
            cp_0['ctlplane_ip'],
            self.outputs['RoleNetIpMap']['Compute']['ctlplane'][0])
        self.assertNotEqual(
            cp_0['ansible_host'],
            self.outputs['RoleNetIpMap']['Compute']['ctlplane'][0])
        # CustomRole
        self.assertNotEqual(
            cs_0['ctlplane_ip'],
            self.outputs['RoleNetIpMap']['CustomRole']['ctlplane'][0])
        self.assertNotEqual(
            cs_0['ansible_host'],
            self.outputs['RoleNetIpMap']['CustomRole']['ctlplane'][0])

        # IP's and hostnames are from neutron while deploy_server_id and
        # bootstrap_server_id, serial etc are from heat.
        expected = {
            'Undercloud': {
                'hosts': {'undercloud': {}},
                'vars': {'ansible_connection': 'local',
                         'ansible_host': 'localhost',
                         'ansible_python_interpreter': sys.executable,
                         'ansible_remote_tmp': '/tmp/ansible-${USER}',
                         'auth_url': 'xyz://keystone.local',
                         'cacert': 'acacert',
                         'overcloud_admin_password': 'theadminpw',
                         'overcloud_keystone_url': 'xyz://keystone',
                         'plan': 'overcloud',
                         'project_name': 'admin',
                         'undercloud_service_list': [
                             'tripleo_nova_compute',
                             'tripleo_heat_engine',
                             'tripleo_ironic_conductor',
                             'tripleo_swift_container_server',
                             'tripleo_swift_object_server',
                             'tripleo_mistral_engine'],
                         'username': 'admin'}},
            'Controller': {
                'hosts': {
                    'c-0': {
                        'ansible_host': '192.0.2.10',
                        'canonical_hostname': 'c-0.example.com.',
                        'ctlplane_hostname': 'c-0.ctlplane.example.com.',
                        'ctlplane_ip': '192.0.2.10',
                        'deploy_server_id': 'a',
                        'internal_api_hostname': 'c-0.internalapi.example.com',
                        'internal_api_ip': '198.51.100.140'},
                    'c-1': {
                        'ansible_host': '192.0.2.11',
                        'canonical_hostname': 'c-1.example.com.',
                        'ctlplane_hostname': 'c-1.ctlplane.example.com.',
                        'ctlplane_ip': '192.0.2.11',
                        'deploy_server_id': 'b',
                        'internal_api_hostname': 'c-1.internalapi.example.com',
                        'internal_api_ip': '198.51.100.141'},
                    'c-2': {
                        'ansible_host': '192.0.2.12',
                        'canonical_hostname': 'c-2.example.com.',
                        'ctlplane_hostname': 'c-2.ctlplane.example.com.',
                        'ctlplane_ip': '192.0.2.12',
                        'deploy_server_id': 'c',
                        'internal_api_hostname': 'c-2.internalapi.example.com',
                        'internal_api_ip': '198.51.100.142'}},
                'vars': {
                    'ansible_ssh_user': 'heat-admin',
                    'bootstrap_server_id': 'a',
                    'ctlplane_cidr': '24',
                    'ctlplane_dns_nameservers': ['192.0.2.253', '192.0.2.254'],
                    'ctlplane_gateway_ip': '192.0.2.1',
                    'ctlplane_host_routes': [{'default': True,
                                              'nexthop': '192.0.2.1'}],
                    'ctlplane_mtu': 1500,
                    'ctlplane_subnet_cidr': '24',
                    'ctlplane_vlan_id': '1',
                    'internal_api_cidr': '25',
                    'internal_api_dns_nameservers': [],
                    'internal_api_gateway_ip': '198.51.100.129',
                    'internal_api_host_routes': [],
                    'internal_api_mtu': 1500,
                    'internal_api_vlan_id': '20',
                    'networks_all': ['InternalApi'],
                    'networks_lower': {'InternalApi': 'internal_api',
                                       'ctlplane': 'ctlplane'},
                    'role_networks': ['ctlplane', 'InternalApi'],
                    'serial': 1,
                    'tripleo_role_name': 'Controller',
                    'tripleo_role_networks': ['ctlplane', 'internal_api']}
            },
            'Compute': {
                'hosts': {
                    'cp-0': {
                        'ansible_host': '192.0.2.20',
                        'canonical_hostname': 'cp-0.example.com.',
                        'ctlplane_hostname': 'cp-0.ctlplane.example.com.',
                        'ctlplane_ip': '192.0.2.20',
                        'deploy_server_id': 'd',
                        'internal_api_hostname':
                            'cp-0.internalapi.example.com',
                        'internal_api_ip': '198.51.100.150'}},
                'vars': {'ansible_ssh_user': 'heat-admin',
                         'bootstrap_server_id': 'a',
                         'ctlplane_cidr': '24',
                         'ctlplane_dns_nameservers': ['192.0.2.253',
                                                      '192.0.2.254'],
                         'ctlplane_gateway_ip': '192.0.2.1',
                         'ctlplane_host_routes': [{'default': True,
                                                   'nexthop': '192.0.2.1'}],
                         'ctlplane_mtu': 1500,
                         'ctlplane_subnet_cidr': '24',
                         'ctlplane_vlan_id': '1',
                         'internal_api_cidr': '25',
                         'internal_api_dns_nameservers': [],
                         'internal_api_gateway_ip': '198.51.100.129',
                         'internal_api_host_routes': [],
                         'internal_api_mtu': 1500,
                         'internal_api_vlan_id': '20',
                         'networks_all': ['InternalApi'],
                         'networks_lower': {'InternalApi': 'internal_api',
                                            'ctlplane': 'ctlplane'},
                         'role_networks': ['ctlplane', 'InternalApi'],
                         'serial': 1,
                         'tripleo_role_name': 'Compute',
                         'tripleo_role_networks': ['ctlplane', 'internal_api']}
            },
            'CustomRole': {
                'hosts': {
                    'cs-0': {
                        'ansible_host': '192.0.2.200',
                        'canonical_hostname': 'cs-0.example.com.',
                        'ctlplane_hostname': 'cs-0.ctlplane.example.com.',
                        'ctlplane_ip': '192.0.2.200',
                        'deploy_server_id': 'e'}},
                'vars': {'ansible_ssh_user': 'heat-admin',
                         'bootstrap_server_id': 'a',
                         'ctlplane_cidr': '24',
                         'ctlplane_dns_nameservers': ['192.0.2.253',
                                                      '192.0.2.254'],
                         'ctlplane_gateway_ip': '192.0.2.1',
                         'ctlplane_host_routes': [{'default': True,
                                                   'nexthop': '192.0.2.1'}],
                         'ctlplane_mtu': 1500,
                         'ctlplane_subnet_cidr': '24',
                         'ctlplane_vlan_id': '1',
                         'internal_api_mtu': 1500,
                         'networks_all': ['InternalApi'],
                         'networks_lower': {'InternalApi': 'internal_api',
                                            'ctlplane': 'ctlplane'},
                         'role_networks': ['ctlplane'],
                         'serial': 1,
                         'tripleo_role_name': 'CustomRole',
                         'tripleo_role_networks': ['ctlplane']}
            },
            'allovercloud': {
                'children': {'Compute': {},
                             'Controller': {},
                             'CustomRole': {}},
                'vars': {'container_cli': 'podman',
                         'ctlplane_vip': 'x.x.x.4',
                         'redis_vip': 'x.x.x.6'}
            },
            'overcloud': {'children': {'allovercloud': {}}},
            'sa': {'children': {'Controller': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}},
            'se': {'children': {'Compute': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}},
            'sd': {'children': {'Compute': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}},
            'sb': {'children': {'Controller': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}},
            'sg': {'children': {'CustomRole': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}},
            'ceph_client': {'children': {'Compute': {}},
                            'vars': {'ansible_ssh_user': 'heat-admin'}},
            'sh': {'children': {'CustomRole': {}},
                   'vars': {'ansible_ssh_user': 'heat-admin'}},
            'clients': {'children': {'Compute': {}},
                        'vars': {'ansible_ssh_user': 'heat-admin'}},
        }
        for k in expected:
            self.assertEqual(expected[k], inv_list[k])

    def test__extend_inventory(self):
        dynamic = False
        existing_inventory = OrderedDict()
        existing_inventory.update({
            'RoleA': {
                'hosts': {
                    'host0': {
                        'existing': 'existing_value'
                    }
                },
                'vars': {
                    'existing': 'existing_value'
                },
            }
        })
        extend_data = {
            'RoleA': {
                'hosts': {
                    'host0': {
                        'new': 'new_value',
                        'existing': 'not_overwritten',
                    }
                },
                'vars': {
                    'new': 'new_var_is_added',
                    'existing': 'not_overwritten',
                },
            }
        }
        expected_inventory = OrderedDict([(
            'RoleA', {
                'hosts': {
                    'host0': {
                        'existing': 'existing_value',
                        'new': 'new_value'
                    }
                },
                'vars': {
                    'existing': 'existing_value',
                    'new': 'new_var_is_added'
                }
            }
        )])

        self.inventory._extend_inventory(existing_inventory, dynamic,
                                         data=extend_data)
        self.assertEqual(expected_inventory, existing_inventory)

    def test__extend_inventory_dynamic(self):
        dynamic = True
        existing_inventory = OrderedDict()
        existing_inventory.update({
            'RoleA': {
                'hosts': {
                    'host0': {
                        'existing': 'existing_value'
                    }
                },
                'vars': {
                    'existing': 'existing_value'
                },
            }
        })
        extend_data = {
            'RoleA': {
                'hosts': {
                    'host0': {
                        'new': 'new_value',
                        'existing': 'not_overwritten',
                    }
                },
                'vars': {
                    'new': 'new_var_is_added',
                    'existing': 'not_overwritten',
                },
            }
        }
        expected_inventory = OrderedDict([(
            'RoleA', {
                'hosts': ['host0'],
                'vars':
                    {'existing': 'existing_value',
                     'new': 'new_var_is_added'}})])

        self.inventory._extend_inventory(existing_inventory, dynamic,
                                         data=extend_data)
        self.assertEqual(expected_inventory, existing_inventory)
        self.assertEqual(
            {'host0': {'existing': 'existing_value',
                       'new': 'new_value'}}, self.inventory.hostvars)


class TestNeutronData(base.TestCase):
    def setUp(self):
        super(TestNeutronData, self).setUp()
        fake_ports = (neutron_fakes.controller0_ports +
                      neutron_fakes.controller1_ports +
                      neutron_fakes.compute_0_ports)
        self.neutron_data = NeutronData(networks=neutron_fakes.fake_networks,
                                        subnets=neutron_fakes.fake_subnets,
                                        ports=fake_ports)

    def test__tags_to_dict(self):
        tags = ['tripleo_foo=foo', 'tripleo_bar=bar', 'other_tag']
        self.assertEqual({'tripleo_foo': 'foo', 'tripleo_bar': 'bar'},
                         NeutronData._tags_to_dict(self, tags))

    def test__networks_by_id(self):
        self.assertEqual({
            'ctlplane_network_id': {
                'dns_domain': 'ctlplane.example.com.',
                'mtu': 1500,
                'name': 'ctlplane',
                'name_upper': 'ctlplane',
                'subnet_ids': ['ctlplane_subnet_id'],
                'tags': {}},
            'internal_api_network_id': {
                'dns_domain': 'internalapi.example.com',
                'mtu': 1500,
                'name': 'internal_api',
                'name_upper': 'InternalApi',
                'subnet_ids': ['internal_api_subnet_id'],
                'tags': {'tripleo_net_idx': 0,
                         'tripleo_network_name': 'InternalApi',
                         'tripleo_vip': True}
            },
        }, self.neutron_data.networks_by_id)

    def test__subnets_by_id(self):
        self.assertEqual({
            'ctlplane_subnet_id': {
                'cidr': '192.0.2.0/24',
                'dns_nameservers': ['192.0.2.253', '192.0.2.254'],
                'gateway_ip': '192.0.2.1',
                'host_routes': [],
                'ip_version': 4,
                'name': 'ctlplane-subnet',
                'network_id': 'ctlplane_network_id',
                'tags': {}
            },
            'internal_api_subnet_id': {
                'cidr': '198.51.100.128/25',
                'dns_nameservers': [],
                'gateway_ip': '198.51.100.129',
                'host_routes': [],
                'ip_version': 4,
                'name': 'internal_api_subnet',
                'network_id': 'internal_api_network_id',
                'tags': {'tripleo_vlan_id': '20'}
            },
        }, self.neutron_data.subnets_by_id)

    def test__ports_by_role_and_host(self):
        self.assertTrue(
            'Controller' in self.neutron_data.ports_by_role_and_host)
        self.assertTrue(
            'Compute' in self.neutron_data.ports_by_role_and_host)
        ctr_role = self.neutron_data.ports_by_role_and_host['Controller']
        cmp_role = self.neutron_data.ports_by_role_and_host['Compute']
        self.assertTrue('c-0' in ctr_role)
        self.assertTrue('c-1' in ctr_role)
        ctr_0 = ctr_role['c-0']
        ctr_1 = ctr_role['c-1']
        self.assertTrue('cp-0' in cmp_role)
        cmp_0 = cmp_role['cp-0']
        self.assertEqual(
            [{'cidr': '24',
              'dns_domain': 'ctlplane.example.com.',
              'dns_nameservers': ['192.0.2.253', '192.0.2.254'],
              'fixed_ips': [{'ip_address': '192.0.2.10',
                             'subnet_id': 'ctlplane_subnet_id'}],
              'gateway_ip': '192.0.2.1',
              'host_routes': [{'default': True, 'nexthop': '192.0.2.1'}],
              'hostname': 'c-0',
              'ip_address': '192.0.2.10',
              'mtu': 1500,
              'name': 'c-0-ctlplane',
              'network_id': 'ctlplane_network_id',
              'network_name': 'ctlplane',
              'subnet_id': 'ctlplane_subnet_id',
              'tags': {'tripleo_default_route': True,
                       'tripleo_network_name': 'ctlplane',
                       'tripleo_role': 'Controller',
                       'tripleo_stack': 'overcloud'},
              'vlan_id': '1'},
             {'cidr': '25',
              'dns_domain': 'internalapi.example.com',
              'dns_nameservers': [],
              'fixed_ips': [{'ip_address': '198.51.100.140',
                             'subnet_id': 'internal_api_subnet_id'}],
              'gateway_ip': '198.51.100.129',
              'host_routes': [],
              'hostname': 'c-0',
              'ip_address': '198.51.100.140',
              'mtu': 1500,
              'name': 'c-0-internal_api',
              'network_id': 'internal_api_network_id',
              'network_name': 'internal_api',
              'subnet_id': 'internal_api_subnet_id',
              'tags': {'tripleo_default_route': False,
                       'tripleo_network_name': 'InternalApi',
                       'tripleo_role': 'Controller',
                       'tripleo_stack': 'overcloud'},
              'vlan_id': '20'}],
            ctr_0
        )
        self.assertEqual(
            [{'cidr': '24',
              'dns_domain': 'ctlplane.example.com.',
              'dns_nameservers': ['192.0.2.253', '192.0.2.254'],
              'fixed_ips': [{'ip_address': '192.0.2.11',
                             'subnet_id': 'ctlplane_subnet_id'}],
              'gateway_ip': '192.0.2.1',
              'host_routes': [{'default': True, 'nexthop': '192.0.2.1'}],
              'hostname': 'c-1',
              'ip_address': '192.0.2.11',
              'mtu': 1500,
              'name': 'c-1-ctlplane',
              'network_id': 'ctlplane_network_id',
              'network_name': 'ctlplane',
              'subnet_id': 'ctlplane_subnet_id',
              'tags': {'tripleo_default_route': True,
                       'tripleo_network_name': 'ctlplane',
                       'tripleo_role': 'Controller',
                       'tripleo_stack': 'overcloud'},
              'vlan_id': '1'},
             {'cidr': '25',
              'dns_domain': 'internalapi.example.com',
              'dns_nameservers': [],
              'fixed_ips': [{'ip_address': '198.51.100.141',
                             'subnet_id': 'internal_api_subnet_id'}],
              'gateway_ip': '198.51.100.129',
              'host_routes': [],
              'hostname': 'c-1',
              'ip_address': '198.51.100.141',
              'mtu': 1500,
              'name': 'c-1-internal_api',
              'network_id': 'internal_api_network_id',
              'network_name': 'internal_api',
              'subnet_id': 'internal_api_subnet_id',
              'tags': {'tripleo_default_route': False,
                       'tripleo_network_name': 'InternalApi',
                       'tripleo_role': 'Controller',
                       'tripleo_stack': 'overcloud'},
              'vlan_id': '20'}],
            ctr_1
        )
        self.assertEqual(
            [{'cidr': '24',
              'dns_domain': 'ctlplane.example.com.',
              'dns_nameservers': ['192.0.2.253', '192.0.2.254'],
              'fixed_ips': [{'ip_address': '192.0.2.20',
                             'subnet_id': 'ctlplane_subnet_id'}],
              'gateway_ip': '192.0.2.1',
              'host_routes': [{'default': True, 'nexthop': '192.0.2.1'}],
              'hostname': 'cp-0',
              'ip_address': '192.0.2.20',
              'mtu': 1500,
              'name': 'cp-0-ctlplane',
              'network_id': 'ctlplane_network_id',
              'network_name': 'ctlplane',
              'subnet_id': 'ctlplane_subnet_id',
              'tags': {'tripleo_default_route': True,
                       'tripleo_network_name': 'ctlplane',
                       'tripleo_role': 'Compute',
                       'tripleo_stack': 'overcloud'},
              'vlan_id': '1'},
             {'cidr': '25',
              'dns_domain': 'internalapi.example.com',
              'dns_nameservers': [],
              'fixed_ips': [{'ip_address': '198.51.100.150',
                             'subnet_id': 'internal_api_subnet_id'}],
              'gateway_ip': '198.51.100.129',
              'host_routes': [],
              'hostname': 'cp-0',
              'ip_address': '198.51.100.150',
              'mtu': 1500,
              'name': 'cp-0-internal_api',
              'network_id': 'internal_api_network_id',
              'network_name': 'internal_api',
              'subnet_id': 'internal_api_subnet_id',
              'tags': {'tripleo_default_route': False,
                       'tripleo_network_name': 'InternalApi',
                       'tripleo_role': 'Compute',
                       'tripleo_stack': 'overcloud'},
              'vlan_id': '20'}],
            cmp_0
        )
        self.assertEqual({'Controller': ctr_role, 'Compute': cmp_role},
                         self.neutron_data.ports_by_role_and_host)



# Copyright (c) 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from tripleo_common import constants
from tripleo_common.tests import base
from tripleo_common.utils import passwords as password_utils
from tripleo_common.utils import plan as plan_utils


PLAN_ENV_CONTENTS = """
version: 1.0

name: overcloud
template: overcloud.yaml
environments:
-  path: overcloud-resource-registry-puppet.yaml
-  path: environments/services/sahara.yaml
parameter_defaults:
  BlockStorageCount: 42
  OvercloudControlFlavor: yummy
passwords:
  AdminPassword: aaaa
  ZaqarPassword: zzzz
"""

USER_ENV_CONTENTS = """
resource_registry:
  OS::TripleO::Foo: bar.yaml
"""

UNORDERED_PLAN_ENV_LIST = [
    {'path': 'overcloud-resource-registry-puppet.yaml'},
    {'path': 'environments/docker-ha.yaml'},
    {'path': 'environments/custom-environment-not-in-capabilities-map.yaml'},
    {'path': 'environments/containers-default-parameters.yaml'},
    {'path': 'environments/docker.yaml'}
]

CAPABILITIES_DICT = {
    'topics': [{
        'environment_groups': [{
            'environments': [{
                'file': 'overcloud-resource-registry-puppet.yaml'}
            ]}, {
            'environments': [{
                'file': 'environments/docker.yaml',
                'requires': ['overcloud-resource-registry-puppet.yaml']
            }, {
                'file': 'environments/containers-default-parameters.yaml',
                'requires': ['overcloud-resource-registry-puppet.yaml',
                             'environments/docker.yaml']
            }]}, {
            'environments': [{
                'file': 'environments/docker-ha.yaml',
                'requires': ['overcloud-resource-registry-puppet.yaml',
                             'environments/docker.yaml']
            }]}
        ]
    }]
}

_EXISTING_PASSWORDS = {
    'PlacementPassword': 'VFJeqBKbatYhQm9jja67hufft',
    'MistralPassword': 'VFJeqBKbatYhQm9jja67hufft',
    'BarbicanPassword': 'MGGQBtgKT7FnywvkcdMwE9nhx',
    'BarbicanSimpleCryptoKek': 'dGhpcnR5X3R3b19ieXRlX2tleWJsYWhibGFoYmxhaGg=',
    'AdminPassword': 'jFmY8FTpvtF2e4d4ReXvmUP8k',
    'CeilometerMeteringSecret': 'CbHTGK4md4Cc8P8ZyzTns6wry',
    'ZaqarPassword': 'bbFgCTFbAH8vf9n3xvZCP8aMR',
    'NovaPassword': '7dZATgVPwD7Ergs9kTTDMCr7F',
    'MysqlRootPassword': 'VqJYpEdKks',
    'RabbitCookie': 'BqJYpEdKksAqJYpEdKks',
    'HeatAuthEncryptionKey': '9xZXehsKc2HbmFFMKjuqxTJHn',
    'PcsdPassword': 'KjEzeitus8eu751a',
    'HorizonSecret': 'mjEzeitus8eu751B',
    'NovajoinPassword': '7dZATgVPwD7Ergs9kTTDMCr7F',
    'IronicPassword': '4hFDgn9ANeVfuqk84pHpD4ksa',
    'RedisPassword': 'xjj3QZDcUQmU6Q7NzWBHRUhGd',
    'SaharaPassword': 'spFvYGezdFwnTk7NPxgYTbUPh',
    'AdminToken': 'jq6G6HyZtj7dcZEvuyhAfjutM',
    'CinderPassword': 'dcxC3xyUcrmvzfrrxpAd3REcm',
    'CongressPassword': 'DwcKvMqXMuNYYFU4zTCuG4234',
    'GlancePassword': 'VqJYNEdKKsGZtgnHct77XBtrV',
    'RabbitPassword': 'ahuHRXdPMx9rzCdjD9CJJNCgA',
    'RpcPassword': 'ahuHRXdPMx9rzCdjD9CJJNCgA',
    'NotifyPassword': 'ahuHRXdPMx9rzCdjD9CJJNCgA',
    'HAProxyStatsPassword': 'P8tbdK6n4YUkTaUyy8XgEVTe6',
    'CeilometerPassword': 'RRdpwK6qf2pbKz2UtzxqauAdk',
    'GnocchiPassword': 'cRYHcUkMuJeK3vyU9pCaznUZc',
    'HeatStackDomainAdminPassword': 'GgTRyWzKYsxK4mReTJ4CM6sMc',
    'CephRgwKey': b'AQCQXtlXAAAAABAAUKcqUMu6oMjAXMjoUV4/3A==',
    'AodhPassword': '8VZXehsKc2HbmFFMKYuqxTJHn',
    'PankoPassword': 'cVZXehsSc2KdmFFMKDudxTLKn',
    'OctaviaHeartbeatKey': 'oct-heartbeat-key',
    'OctaviaPassword': 'NMl7j3nKk1VVwMxUZC8Cgw==',
    'OctaviaServerCertsKeyPassphrase': 'aW5zZWN1cmUta2V5LWRvLW5vdC11c2U=',
    'OctaviaCaKeyPassphrase': 'SLj4c3uCk4DDxPwQOG1Heb==',
    'ManilaPassword': 'NYJN86Fua3X8AVFWmMhQa2zTH',
    'NeutronMetadataProxySharedSecret': 'Q2YgUCwmBkYdqsdhhCF4hbghu',
    'CephManilaClientKey': b'AQANOFFY1NW6AxAAu6jWI3YSOsp2QWusb5Y3DQ==',
    'CephGrafanaAdminPassword': 'NYJN86Fua3X8AVFWmMhQa2zTH',
    'CephDashboardAdminPassword': 'NYJN86Fua3X8AVFWmMhQa2zTH',
    'SwiftHashSuffix': 'td8mV6k7TYEGKCDvjVBwckpn9',
    'SnmpdReadonlyUserPassword': 'TestPassword',
    'SwiftPassword': 'z6EWAVfW7CuxvKdzjWTdrXCeg',
    'HeatPassword': 'bREnsXtMHKTHxt8XW6NXAYr48',
    'MysqlClustercheckPassword': 'jN4RMMWWJ4sycaRwh7UvrAtfX',
    'CephClientKey': b'AQCQXtlXAAAAABAAKyc+8St8i9onHyu2mPk+vg==',
    'NeutronPassword': 'ZxAjdU2UXCV4GM3WyPKrzAZXD',
    'DesignatePassword': 'wHYj7rftFzHMpJKnGxbjjR9CW',
    'DesignateRndcKey': 'hB8XaZRd2Tf00jKsyoXpyw==',
    'KeystoneCredential0': 'ftJNQ_XlDUK7Lgvv1kdWf3SyqVsrvNDgoNV4kJg3yzw=',
    'KeystoneCredential1': 'c4MFq82TQLFLKpiiUjrKkp15dafE2ALcD3jbaIu3rfE=',
    'KeystoneFernetKey0': 'O8NSPxr4zXBBAoGIj-5aUmtE7-Jk5a4ptVsEhzJ8Vd8=',
    'KeystoneFernetKey1': 'AueoL37kd6eLjV29AG-Ruxu5szW47osgXx6aPOqtI6I=',
    'KeystoneFernetKeys': {
        '/etc/keystone/fernet-keys/0': {'content': 'IAMAVERYSAFEKEY'},
        '/etc/keystone/fernet-keys/1': {'content': 'IALSOAMAVERYSAFEKEY'}
    },
    'CephClusterFSID': u'97c16f44-b62c-11e6-aed3-185e0f73fdc5',
    'Ec2ApiPassword': 'FPvz2WiWxrHVWrmSSvv44bqmr',
    'EtcdInitialClusterToken': 'fcVZXehsSc2KdmFFMKDudxTLKa',
    'PacemakerRemoteAuthkey':
        'bCfHQx4fX7FqENVBbDfBnKvf6FTH6mPfVdNjfzakEjuF4UbmZJHAxWdheEr6feEyZmtM'
        'XEd4w3qM8nMVrzjnDCmqAFDmMDQfKcuNgTnqGnkbVUDGpym67Ry4vNCPHyp9tGGyfjNX'
        't66csYZTYUHPv6jdJk4HWBjE66v8B3nRpc3FePQ8DRMWX4hcGFNNxapJu7v2frKwq4tD'
        '78cc7aPPMGPn8kR3mj7kMP8Ah8VVGXJEtybEvRg4sQ67zEkAzfKggrpXYPK2Qvv9sHKp'
        't2VjwZBHTvWKarJjyeMTqbzJyW6JTbm62gqZCr9afZRFQug62pPRduvkUNfUYNPNpqjy'
        'yznmeAZPxVseU3jJVxKrxdrgzavKEMtW6BbTmw86j8wuUdaWgRccRGVUQvtQ4p9kXHAy'
        'eXVduZvpvxFtbKvfNTvf6qCuJ8qeQp2TwJQPHUYHkxZYrpAA7fZUzNCZR2tFFdZzWGt2'
        'PEnYvYts4m7Fp9XEmNm7Jyme38CBfnaVERmTMRvHkq3EE2Amsc72aDdzeVRjR3xRgMNJ'
        '2cEEWqatZXveHxJr6VmBNWJUyvPrfmVegwtKCGJND8d3Ysruy7GCn6zcrNY7d84aDk3P'
        'q7NyZfRYrGcNDKJuzNWH8UNwGP68uQsUUrV9NVTVpB2sRPG2tJm3unYqekUg3KYXu46J'
        'mANxqgrqDv6vPx6NCPdUXZTXFaesQatKRkkf3nZFqZQJXZVbkudTmrPYyRQAjvWuAmrY'
        '6RcFFmygeFnhAxhwXNdge9tEfsfPeQ4GMxa8Amj2fMjmNvQXFfQ8uxMUnusDmhbwCRKM'
        'CvN2dNE92MaQge34vtxsueyDEmbuVE9sNRD3EQBRwx8nktgRwKHfRZJ3BX8f9XMaQe2e'
        'ZfGjtUNkbgKdCyYgEwEybXKPfevDnxFvbZMpJx4fqqCAbAZud9RnAuvqHgFbKHXcVEE4'
        'nRmgJmdqJsRsTkYPpYkKN9rssEDCXr9HFjbenkxXcUe8afrTvKAzwBvbDWcjYBEQKbuY'
        '6Ptm9VJrjutUHCPmW2sh66qvq4C9vPhVEey7FpCZDEyYUPrjRfhKjxEFNBKWpcZzvmT2'
        'nRmgJmdqJsRsTkYPpYkKN9rssEDCXr9HFjbenkxXcUe8afrTvKAzwBvbDWcjYBEQKbuY'
        '2cEEWqatZXveHxJr6VmBNWJUyvPrfmVegwtKCGJND8d3Ysruy7GCn6zcrNY7d84aDk3P'
        'VRE4aqMfuY72xFacxXHjvWagEGQEYtkMtQnsh7XAMGuazT3pkppeUTyDbKTY2Dz7Quc3'
        '8UKaw8ece6fTXWpjX2EYrsd4qzvhC6eEPdgnpmzjqmuG8YqEAUZ7dYADgAhTkBQsNct8'
        'btQsQDYD4PBjxG2KWAZ9vgTsvBpjjEVcrPfWgwZKJTAZWfWq2u7nT4N2t39EYmQEzbEf'
        '8UKaw8ece6fTXWpjX2EYrsd4qzvhC6eEPdgnpmzjqmuG8YqEAUZ7dYADgAhTkBQsNct8'
        'DkCF3DJ49jjZm9N4EKnKGGXD7XkFE79AFRGPUw4gXpeQCtUXyEugUErqMjqgJjC7ykdg'
        'zz7txnzYfRaKHNVs4r4GwNEHRHt7VcTuT3WBcbE4skQgjMnttgP7hts7dMU7PA8kRrfq'
        'BKdkPkUwqQ9Xn4zrysY4GvJQHWXxD6Tyqf9PZaz4xbUmsvtuY7NAz27U2aT3EA9XCgfn'
        '2cEEWqatZXveHxJr6VmBNWJUyvPrfmVegwtKCGJND8d3Ysruy7GCn6zcrNY7d84aDk3P'
        'CEfTJQz342nwRMY4DCuhawz4cnrWwxgsnVPCbeXYH4RcgswVsk9edxKkYMkpTwpcKf6n'
        'nRmgJmdqJsRsTkYPpYkKN9rssEDCXr9HFjbenkxXcUe8afrTvKAzwBvbDWcjYBEQKbuY'
        '6Ptm9VJrjutUHCPmW2sh66qvq4C9vPhVEey7FpCZDEyYUPrjRfhKjxEFNBKWpcZzvmT2'
        'VRE4aqMfuY72xFacxXHjvWagEGQEYtkMtQnsh7XAMGuazT3pkppeUTyDbKTY2Dz7Quc3'
        '8UKaw8ece6fTXWpjX2EYrsd4qzvhC6eEPdgnpmzjqmuG8YqEAUZ7dYADgAhTkBQsNct8'
        'btQsQDYD4PBjxG2KWAZ9vgTsvBpjjEVcrPfWgwZKJTAZWfWq2u7nT4N2t39EYmQEzbEf'
        'DkCF3DJ49jjZm9N4EKnKGGXD7XkFE79AFRGPUw4gXpeQCtUXyEugUErqMjqgJjC7ykdg'
        'zz7txnzYfRaKHNVs4r4GwNEHRHt7VcTuT3WBcbE4skQgjMnttgP7hts7dMU7PA8kRrfq'
        'BKdkPkUwqQ9Xn4zrysY4GvJQHWXxD6Tyqf9PZaz4xbUmsvtuY7NAz27U2aT3EA9XCgfn'
        '2cEEWqatZXveHxJr6VmBNWJUyvPrfmVegwtKCGJND8d3Ysruy7GCn6zcrNY7d84aDk3P'
        'CEfTJQz342nwRMY4DCuhawz4cnrWwxgsnVPCbeXYH4RcgswVsk9edxKkYMkpTwpcKf6n'
        'E2dhquqdKVTAYf7YKbTfFVsRwqykkPduKXuPwVDjbCqdEJPcmnRJAJkwkQCWgukpvzzm'
        'DKFVYxncxmzKgEN27VtgfpsXWBJ2jaxMeQCXb2rbjkVcaypyaETQ3Wkw98EptNAKRcjM'
        'E2dhquqdKVTAYf7YKbTfFVsRwqykkPduKXuPwVDjbCqdEJPcmnRJAJkwkQCWgukpvzzm'
        'zZJ2xFdfNYh7RZ7EgAAbY8Tqy3j2c9c6HNmXwAVV6dzPTrE4FHcKZGg76anGchczF9ev'
        'AG8RHQ7ea2sJhXqBmGsmEj6Q84TN9E7pgmtAtmVAA38AYsQBNZUMYdMcmBdpV9w7G3NZ'
        'mEU8R8uWqx6w3NzzqsMg78bnhCR7sdWDkhuEp2M8fYWmqujYFNYvzz6BcHNKQyrWETRD'
        'E2dhquqdKVTAYf7YKbTfFVsRwqykkPduKXuPwVDjbCqdEJPcmnRJAJkwkQCWgukpvzzm'
        'zaTdNWgM7wsXGkvgYVNdTWnReCPXJUN3yQwrvApZzdaF86QaeYwXW7qqEJrqmwpUUbw2'
        'JHkmvJB4AWtVhDc9etzUqfuTaqMyXwxFEWvht3RDTDx8dfQ3Ek8BD4QP4BtUQeQJpfsG'
        'FEJeQQYVcBxqVuK26xJrERUDmeNw8KWKBCrYPPy48cjCFdgZHz3cNet6bwJMdsgKMpZT'
        'erdYy9nqBw6FRZ37rRMtxmrcB4VsWHbf4HjdPRpu4xyJTqMThnXWa8nPDde3C9wCuKkQ'
        '23k2zDYsMeHc6KD93vm7Ky48v3veYEuJvNNxQPyyCZ9XNnpGsWrqsVduCswR4MQpp6yJ'
        'RBmwbMYbuEjwJy9UuZxa9bQV4GqYFnVuETC6bXaT9uauWdaa2TrbuuXx3WWdmRGd4Rqh'
        'Z3NA9Kqx9pTQHe3KGZ2tFejsJqNvjJvFX94eVeMGDgHjtJzDdxp9NWYtG6v9zABGRzVF'
        'MqJX6nhhBPbsvjpswcgJq3ZXxzmWFJmvjECghGrbG6bKawtv4aYhMeaHagfMP8W6KrTy'
        'uGxWUhcEhfygjE4truAkjfKCtzzVtTcBArbWMny6HWMp6TAen3f6hEB6kBb7pgvKxkND'
        '3JxueYBZvDeq4WWtRzUjcFF2qhEjwrtuCJhy3WMXX3MN6nFDtYRTHZGdPqyatW9Jcc8t'
        '7gCMWMVzYyNuXZ2A6rwX6Umv8g3mBuwnrwKXEFTZkPCAZMxk3A6MTmMcJCVy3hw6MmRM'
        'eXKyhFxRcKWraysTQG7hd9kP8DeJZNDurYDJwqrh6cwDwaMhBfTgnxTBeyjwpbCJK2FD'
        'Jg2vFWPmTJ37gDMdwxWCMRQ9kyqz9PJZ4Xn2MPxMhNqT3Hb39YshryqnbvBagHbqYx9M'
        'r4ZKJpKya34JMaPambzg2pKRDd2WdFCZcdHTFyqxxzJbjXM2gjfBZ2strUNqWvQYNTw8'
        'QttkuxyeQTgHupKNaZF6y7rDyf7mbNR9DaPXpBQuZ7un6KDj2Dfh7yvfhPk8cHG7n9pb'
        'KEKD3sgbbKnQ8d9MsGhUtCQVed7dtjpYKsmGJmbYMvZjpGpqsfsHQfFRdCgJHnW3FdQ6'
        'sGhUtCQVed7dtj12',
    'MigrationSshKey': {
        'private_key': 'private_key',
        'public_key': 'public_key'
        },
    'LibvirtTLSPassword': 'xCdt9yeamKz8Fb6EGba9u82XU',
}


class PlanTest(base.TestCase):
    def setUp(self):
        super(PlanTest, self).setUp()
        self.container = 'overcloud'

    def test_get_next_index(self):
        keys_map = {
            password_utils.KEYSTONE_FERNET_REPO + '0': {
                'content': 'Some key'},
            password_utils.KEYSTONE_FERNET_REPO + '1': {
                'content': 'Some other key'},
        }
        next_index = plan_utils.get_next_index(keys_map)
        self.assertEqual(next_index, 2)

    @mock.patch('tripleo_common.utils.passwords.'
                'create_keystone_credential')
    def test_rotate_keys(self, mock_keystone_creds):
        mock_keystone_creds.return_value = 'Some new key'

        staged_key_index = password_utils.KEYSTONE_FERNET_REPO + '0'
        new_primary_key_index = password_utils.KEYSTONE_FERNET_REPO + '2'
        keys_map = {
            password_utils.KEYSTONE_FERNET_REPO + '0': {
                'content': 'Some key'},
            password_utils.KEYSTONE_FERNET_REPO + '1': {
                'content': 'Some other key'},
        }
        new_keys_map = plan_utils.rotate_keys(keys_map, 2)

        # Staged key should be the new key
        self.assertEqual('Some new key',
                         new_keys_map[staged_key_index]['content'])
        # primary key should be the previous staged key
        self.assertEqual('Some key',
                         new_keys_map[new_primary_key_index]['content'])

    def test_purge_excess_keys_should_purge(self):
        keys_map = {
            password_utils.KEYSTONE_FERNET_REPO + '0': {
                'content': 'key0'},
            password_utils.KEYSTONE_FERNET_REPO + '1': {
                'content': 'key1'},
            password_utils.KEYSTONE_FERNET_REPO + '2': {
                'content': 'key2'},
            password_utils.KEYSTONE_FERNET_REPO + '3': {
                'content': 'key3'},
            password_utils.KEYSTONE_FERNET_REPO + '4': {
                'content': 'key4'},
        }
        max_keys = 3
        keys_map = plan_utils.purge_excess_keys(max_keys, keys_map)
        self.assertEqual(max_keys, len(keys_map))
        # It should keep index 0, 3 and 4
        self.assertIn(password_utils.KEYSTONE_FERNET_REPO + '0', keys_map)
        self.assertIn(password_utils.KEYSTONE_FERNET_REPO + '3', keys_map)
        self.assertIn(password_utils.KEYSTONE_FERNET_REPO + '4', keys_map)
        # It sould have removed index 1 and 2
        self.assertNotIn(password_utils.KEYSTONE_FERNET_REPO + '1', keys_map)
        self.assertNotIn(password_utils.KEYSTONE_FERNET_REPO + '2', keys_map)

    def test_purge_excess_keys_should_not_purge_if_equal_to_max(self):
        keys_map = {
            password_utils.KEYSTONE_FERNET_REPO + '0': {
                'content': 'key0'},
            password_utils.KEYSTONE_FERNET_REPO + '1': {
                'content': 'key1'},
            password_utils.KEYSTONE_FERNET_REPO + '2': {
                'content': 'key2'},
        }
        max_keys = 3
        keys_map = plan_utils.purge_excess_keys(max_keys, keys_map)
        self.assertEqual(max_keys, len(keys_map))

    def test_purge_excess_keys_should_not_purge_if_less_than_max(self):
        keys_map = {
            password_utils.KEYSTONE_FERNET_REPO + '0': {
                'content': 'key0'},
            password_utils.KEYSTONE_FERNET_REPO + '1': {
                'content': 'key1'},
        }
        max_keys = 3
        keys_map = plan_utils.purge_excess_keys(max_keys, keys_map)
        self.assertEqual(2, len(keys_map))

    @mock.patch('tripleo_common.utils.passwords.'
                'get_snmpd_readonly_user_password')
    def test_generate_password(self, mock_get_snmpd_readonly_user_password):

        mock_get_snmpd_readonly_user_password.return_value = "TestPassword"

        mock_orchestration = mock.MagicMock()
        mock_orchestration.stacks.environment.return_value = {
            'parameter_defaults': {}
        }
        mock_resource = mock.MagicMock()
        mock_resource.attributes = {
            'endpoint_map': {
                'PlacementPublic': {}
            },
        }
        mock_orchestration.resources.get.return_value = mock_resource
        result = plan_utils.generate_passwords(None, mock_orchestration)

        for password_param_name in constants.PASSWORD_PARAMETER_NAMES:
            self.assertTrue(password_param_name in result,
                            "%s is not in %s" % (password_param_name, result))

    @mock.patch('tripleo_common.utils.passwords.'
                'create_ssh_keypair')
    @mock.patch('tripleo_common.utils.passwords.'
                'create_fernet_keys_repo_structure_and_keys')
    @mock.patch('tripleo_common.utils.passwords.'
                'get_snmpd_readonly_user_password')
    def test_run_passwords_exist(self, mock_get_snmpd_readonly_user_password,
                                 mock_fernet_keys_setup,
                                 mock_create_ssh_keypair):

        mock_get_snmpd_readonly_user_password.return_value = "TestPassword"
        mock_create_ssh_keypair.return_value = {'public_key': 'Foo',
                                                'private_key': 'Bar'}
        mock_fernet_keys_setup.return_value = {'/tmp/foo': {'content': 'Foo'},
                                               '/tmp/bar': {'content': 'Bar'}}

        mock_orchestration = mock.MagicMock()
        mock_orchestration.stacks.environment.return_value = {
            'parameter_defaults': _EXISTING_PASSWORDS.copy()
        }
        mock_resource = mock.MagicMock()
        mock_resource.attributes = {
            'endpoint_map': {
                'PlacementPublic': {}
            },
        }
        mock_orchestration.resources.get.return_value = mock_resource

        result = plan_utils.generate_passwords(None, mock_orchestration)

        # ensure old passwords used and no new generation
        self.assertEqual(_EXISTING_PASSWORDS, result)

    @mock.patch('tripleo_common.utils.passwords.'
                'create_ssh_keypair')
    @mock.patch('tripleo_common.utils.passwords.'
                'create_fernet_keys_repo_structure_and_keys')
    @mock.patch('tripleo_common.utils.passwords.'
                'get_snmpd_readonly_user_password')
    def test_placement_passwords_upgrade(self,
                                         mock_get_snmpd_readonly_user_password,
                                         mock_fernet_keys_setup,
                                         mock_create_ssh_keypair):

        mock_get_snmpd_readonly_user_password.return_value = "TestPassword"
        mock_create_ssh_keypair.return_value = {'public_key': 'Foo',
                                                'private_key': 'Bar'}
        mock_fernet_keys_setup.return_value = {'/tmp/foo': {'content': 'Foo'},
                                               '/tmp/bar': {'content': 'Bar'}}

        passwords = _EXISTING_PASSWORDS.copy()

        mock_orchestration = mock.MagicMock()
        mock_orchestration.stacks.environment.return_value = {
            'parameter_defaults': passwords
        }
        mock_resource = mock.MagicMock()
        mock_resource.attributes = {
            'endpoint_map': {},
        }
        mock_orchestration.resources.get.return_value = mock_resource
        result = plan_utils.generate_passwords(None, mock_orchestration)
        self.assertEqual(
            passwords['NovaPassword'],
            result['PlacementPassword']
        )

    @mock.patch('tripleo_common.utils.passwords.'
                'create_ssh_keypair')
    @mock.patch('tripleo_common.utils.passwords.'
                'create_fernet_keys_repo_structure_and_keys')
    @mock.patch('tripleo_common.utils.passwords.'
                'get_snmpd_readonly_user_password')
    def test_run_rotate_no_rotate_list(
        self, mock_get_snmpd_readonly_user_password,
        mock_fernet_keys_setup, mock_create_ssh_keypair):

        mock_get_snmpd_readonly_user_password.return_value = "TestPassword"
        mock_create_ssh_keypair.return_value = {'public_key': 'Foo',
                                                'private_key': 'Bar'}
        mock_fernet_keys_setup.return_value = {'/tmp/foo': {'content': 'Foo'},
                                               '/tmp/bar': {'content': 'Bar'}}
        mock_orchestration = mock.MagicMock()
        mock_orchestration.stacks.environment.return_value = {
            'parameter_defaults': _EXISTING_PASSWORDS.copy()
        }

        mock_resource = mock.MagicMock()
        mock_resource.attributes = {
            'endpoint_map': {
                'PlacementPublic': {}
            },
        }
        mock_orchestration.resources.get.return_value = mock_resource

        result = plan_utils.generate_passwords(None, mock_orchestration,
                                               rotate_passwords=True)

        # ensure passwords in the DO_NOT_ROTATE_LIST are not modified
        for name in constants.DO_NOT_ROTATE_LIST:
            self.assertEqual(_EXISTING_PASSWORDS[name], result[name])

        # ensure all passwords are generated
        for name in constants.PASSWORD_PARAMETER_NAMES:
            self.assertTrue(name in result, "%s is not in %s" % (name, result))

        # ensure new passwords have been generated
        self.assertNotEqual(_EXISTING_PASSWORDS, result)

    @mock.patch('tripleo_common.utils.passwords.'
                'create_ssh_keypair')
    @mock.patch('tripleo_common.utils.passwords.'
                'create_fernet_keys_repo_structure_and_keys')
    @mock.patch('tripleo_common.utils.passwords.'
                'get_snmpd_readonly_user_password')
    def test_run_rotate_with_rotate_list(
        self, mock_get_snmpd_readonly_user_password,
        mock_fernet_keys_setup, mock_create_ssh_keypair):

        mock_get_snmpd_readonly_user_password.return_value = "TestPassword"
        mock_create_ssh_keypair.return_value = {'public_key': 'Foo',
                                                'private_key': 'Bar'}
        mock_fernet_keys_setup.return_value = {'/tmp/foo': {'content': 'Foo'},
                                               '/tmp/bar': {'content': 'Bar'}}

        mock_orchestration = mock.MagicMock()
        mock_orchestration.stacks.environment.return_value = {
            'parameter_defaults': _EXISTING_PASSWORDS.copy()
        }

        mock_resource = mock.MagicMock()
        mock_resource.attributes = {
            'endpoint_map': {
                'PlacementPublic': {}
            },
        }
        mock_orchestration.resources.get.return_value = mock_resource
        rotate_list = [
            'MistralPassword',
            'BarbicanPassword',
            'AdminPassword',
            'CeilometerMeteringSecret',
            'ZaqarPassword',
            'NovaPassword',
            'MysqlRootPassword'
        ]

        result = plan_utils.generate_passwords(None, mock_orchestration,
                                               rotate_passwords=True,
                                               rotate_pw_list=rotate_list)

        # ensure only specified passwords are regenerated
        for name in constants.PASSWORD_PARAMETER_NAMES:
            self.assertTrue(name in result, "%s is not in %s" % (name, result))
            if name in rotate_list:
                self.assertNotEqual(_EXISTING_PASSWORDS[name], result[name])
            else:
                self.assertEqual(_EXISTING_PASSWORDS[name], result[name])

    @mock.patch('tripleo_common.utils.passwords.'
                'create_ssh_keypair')
    @mock.patch('tripleo_common.utils.passwords.'
                'create_fernet_keys_repo_structure_and_keys')
    @mock.patch('tripleo_common.utils.passwords.'
                'get_snmpd_readonly_user_password')
    def test_passwords_exist_in_heat(
        self, mock_get_snmpd_readonly_user_password,
        mock_fernet_keys_setup, mock_create_ssh_keypair):

        mock_get_snmpd_readonly_user_password.return_value = "TestPassword"
        mock_create_ssh_keypair.return_value = {'public_key': 'Foo',
                                                'private_key': 'Bar'}
        mock_fernet_keys_setup.return_value = {'/tmp/foo': {'content': 'Foo'},
                                               '/tmp/bar': {'content': 'Bar'}}

        existing_passwords = _EXISTING_PASSWORDS.copy()
        existing_passwords["AdminPassword"] = 'ExistingPasswordInHeat'

        mock_orchestration = mock.MagicMock()
        mock_orchestration.stacks.environment.return_value = {
            'parameter_defaults': existing_passwords
        }

        mock_resource = mock.MagicMock()
        mock_resource.attributes = {
            'endpoint_map': {
                'PlacementPublic': {}
            },
        }
        mock_orchestration.resources.get.return_value = mock_resource

        result = plan_utils.generate_passwords(None, mock_orchestration)
        self.assertEqual(existing_passwords, result)



# Copyright (c) 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from osc_lib import utils as osc_lib_utils


LOG = logging.getLogger(__name__)
heatclient = None


class EphemeralHeatClient(object):
    """A Heat client shim class to be used with ephemeral Heat.

    When the heat client is used to talk to the Heat API, the environment will
    be set with the correct variable configuration to configure Keystone for
    auth type none and to use a direct endpoint.

    After the client is finished, the environment is restored. This is
    necessary so that the entire system environment is not reconfigured for
    auth_type=none for the duration of the tripleoclient execution.

    :param heat: Heat client
    :type heat: `heatclient.heatclient`

    """

    def __init__(self, heat, host, port):
        self.heat = heat
        self.host = host
        self.port = port
        os.environ['OS_HEAT_TYPE'] = 'ephemeral'
        os.environ['OS_HEAT_HOST'] = host
        os.environ['OS_HEAT_PORT'] = str(port)

    def save_environment(self):
        self.environ = os.environ.copy()
        for v in ('OS_USER_DOMAIN_NAME',
                  'OS_PROJECT_DOMAIN_NAME',
                  'OS_PROJECT_NAME'):
            os.environ.pop(v, None)

        os.environ['OS_AUTH_TYPE'] = "none"
        os.environ['OS_ENDPOINT'] = self.heat.http_client.endpoint

    def restore_environment(self):
        os.environ = self.environ.copy()

    def __getattr__(self, attr):
        self.save_environment()
        try:
            val = getattr(self.heat, attr)
        finally:
            self.restore_environment()
        return val


def local_orchestration_client(host="127.0.0.1", api_port=8006):
    """Returns a local orchestration service client"""

    API_VERSIONS = {
        '1': 'heatclient.v1.client.Client',
    }

    heat_client = osc_lib_utils.get_client_class(
        'tripleoclient',
        '1',
        API_VERSIONS)
    LOG.debug('Instantiating local_orchestration client for '
              'host %s, port %s: %s',
              host, api_port, heat_client)

    endpoint = 'http://%s:%s/v1/admin' % (host, api_port)
    client = heat_client(
        endpoint=endpoint,
        username='admin',
        password='fake',
        region_name='regionOne',
        token='fake',
    )

    global heatclient
    heatclient = EphemeralHeatClient(client, host, api_port)
    return heatclient



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

class FakeNeutronNetwork(dict):
    def __init__(self, **attrs):
        NETWORK_ATTRS = ['id',
                         'name',
                         'status',
                         'tenant_id',
                         'is_admin_state_up',
                         'mtu',
                         'segments',
                         'is_shared',
                         'subnets',
                         'provider:network_type',
                         'provider:physical_network',
                         'provider:segmentation_id',
                         'router:external',
                         'availability_zones',
                         'availability_zone_hints',
                         'is_default',
                         'tags']

        raw = dict.fromkeys(NETWORK_ATTRS)
        raw.update(attrs)
        raw.update({
            'provider_physical_network': attrs.get(
                'provider:physical_network', None),
            'provider_network_type': attrs.get(
                'provider:network_type', None),
            'provider_segmentation_id': attrs.get(
                'provider:segmentation_id', None)
        })
        super(FakeNeutronNetwork, self).__init__(raw)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        if key in self:
            self[key] = value
        else:
            raise AttributeError(key)


class FakeNeutronPort(dict):
    def __init__(self, **attrs):
        PORT_ATTRS = ['admin_state_up',
                      'allowed_address_pairs',
                      'binding:host_id',
                      'binding:profile',
                      'binding:vif_details',
                      'binding:vif_type',
                      'binding:vnic_type',
                      'data_plane_status',
                      'description',
                      'device_id',
                      'device_owner',
                      'dns_assignment',
                      'dns_domain',
                      'dns_name',
                      'extra_dhcp_opts',
                      'fixed_ips',
                      'id',
                      'mac_address',
                      'name', 'network_id',
                      'port_security_enabled',
                      'security_group_ids',
                      'status',
                      'tenant_id',
                      'qos_network_policy_id',
                      'qos_policy_id',
                      'tags',
                      'uplink_status_propagation']

        raw = dict.fromkeys(PORT_ATTRS)
        raw.update(attrs)
        super(FakeNeutronPort, self).__init__(raw)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        if key in self:
            self[key] = value
        else:
            raise AttributeError(key)


class FakeNeutronSubnet(dict):
    def __init__(self, **attrs):
        SUBNET_ATTRS = ['id',
                        'name',
                        'network_id',
                        'cidr',
                        'tenant_id',
                        'is_dhcp_enabled',
                        'dns_nameservers',
                        'allocation_pools',
                        'host_routes',
                        'ip_version',
                        'gateway_ip',
                        'ipv6_address_mode',
                        'ipv6_ra_mode',
                        'subnetpool_id',
                        'segment_id',
                        'tags']

        raw = dict.fromkeys(SUBNET_ATTRS)
        raw.update(attrs)
        super(FakeNeutronSubnet, self).__init__(raw)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        if key in self:
            self[key] = value
        else:
            raise AttributeError(key)



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tripleo_common.tests.fake_neutron import stubs


ctlplane_network = stubs.FakeNeutronNetwork(
    name='ctlplane',
    id='ctlplane_network_id',
    mtu=1500,
    dns_domain='ctlplane.example.com.',
    subnet_ids=['ctlplane_subnet_id'],
    tags=[],
)
internal_api_network = stubs.FakeNeutronNetwork(
    name='internal_api',
    id='internal_api_network_id',
    mtu=1500,
    dns_domain='internalapi.example.com',
    subnet_ids=['internal_api_subnet_id'],
    tags=['tripleo_net_idx=0',
          'tripleo_vip=true',
          'tripleo_network_name=InternalApi'],
)

ctlplane_subnet = stubs.FakeNeutronSubnet(
    name='ctlplane-subnet',
    id='ctlplane_subnet_id',
    network_id='ctlplane_network_id',
    cidr='192.0.2.0/24',
    gateway_ip='192.0.2.1',
    dns_nameservers=['192.0.2.253', '192.0.2.254'],
    host_routes=[],
    ip_version=4,
    tags=[],
)
internal_api_subnet = stubs.FakeNeutronSubnet(
    name='internal_api_subnet',
    id='internal_api_subnet_id',
    network_id='internal_api_network_id',
    cidr='198.51.100.128/25',
    gateway_ip='198.51.100.129',
    dns_nameservers=[],
    host_routes=[],
    ip_version=4,
    tags=['tripleo_vlan_id=20'],
)


fake_networks = [ctlplane_network, internal_api_network]
fake_subnets = [ctlplane_subnet, internal_api_subnet]

controller0_ports = [
    stubs.FakeNeutronPort(name='c-0-ctlplane',
                          id='controller_0_ctlplane_id',
                          network_id=ctlplane_network.id,
                          fixed_ips=[dict(ip_address='192.0.2.10',
                                          subnet_id=ctlplane_subnet.id)],
                          dns_name='c-0',
                          tags=['tripleo_network_name=ctlplane',
                                'tripleo_role=Controller',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=True'],
                          ),
    stubs.FakeNeutronPort(name='c-0-internal_api',
                          id='controller_0_internal_api_id',
                          network_id=internal_api_network.id,
                          fixed_ips=[dict(ip_address='198.51.100.140',
                                          subnet_id=internal_api_subnet.id)],
                          dns_name='c-0',
                          tags=['tripleo_network_name=InternalApi',
                                'tripleo_role=Controller',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=False'],
                          ),
]

controller1_ports = [
    stubs.FakeNeutronPort(name='c-1-ctlplane',
                          id='controller_1_ctlplane_id',
                          network_id=ctlplane_network.id,
                          fixed_ips=[dict(ip_address='192.0.2.11',
                                          subnet_id=ctlplane_subnet.id)],
                          dns_name='c-1',
                          tags=['tripleo_network_name=ctlplane',
                                'tripleo_role=Controller',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=True'],
                          ),
    stubs.FakeNeutronPort(name='c-1-internal_api',
                          id='controller_1_internal_api_id',
                          network_id=internal_api_network.id,
                          fixed_ips=[dict(ip_address='198.51.100.141',
                                          subnet_id=internal_api_subnet.id)],
                          dns_name='c-1',
                          tags=['tripleo_network_name=InternalApi',
                                'tripleo_role=Controller',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=False'],
                          ),
]

controller2_ports = [
    stubs.FakeNeutronPort(name='c-2-ctlplane',
                          id='controller_2_ctlplane_id',
                          network_id=ctlplane_network.id,
                          fixed_ips=[dict(ip_address='192.0.2.12',
                                          subnet_id=ctlplane_subnet.id)],
                          dns_name='c-2',
                          tags=['tripleo_network_name=ctlplane',
                                'tripleo_role=Controller',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=True'],
                          ),
    stubs.FakeNeutronPort(name='c-2-internal_api',
                          id='controller_2_internal_api_id',
                          network_id=internal_api_network.id,
                          fixed_ips=[dict(ip_address='198.51.100.142',
                                          subnet_id=internal_api_subnet.id)],
                          dns_name='c-2',
                          tags=['tripleo_network_name=InternalApi',
                                'tripleo_role=Controller',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=False'],
                          ),
]

compute_0_ports = [
    stubs.FakeNeutronPort(name='cp-0-ctlplane',
                          id='compute_0_ctlplane_id',
                          network_id=ctlplane_network.id,
                          fixed_ips=[dict(ip_address='192.0.2.20',
                                          subnet_id=ctlplane_subnet.id)],
                          dns_name='cp-0',
                          tags=['tripleo_network_name=ctlplane',
                                'tripleo_role=Compute',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=True'],
                          ),
    stubs.FakeNeutronPort(name='cp-0-internal_api',
                          id='compute_0_internal_api_id',
                          network_id=internal_api_network.id,
                          fixed_ips=[dict(ip_address='198.51.100.150',
                                          subnet_id=internal_api_subnet.id)],
                          dns_name='cp-0',
                          tags=['tripleo_network_name=InternalApi',
                                'tripleo_role=Compute',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=False'],
                          ),

]

custom_0_ports = [
    stubs.FakeNeutronPort(name='cs-0-ctlplane',
                          id='custom_0_ctlplane_id',
                          network_id=ctlplane_network.id,
                          fixed_ips=[dict(ip_address='192.0.2.200',
                                          subnet_id=ctlplane_subnet.id)],
                          dns_name='cs-0',
                          tags=['tripleo_network_name=ctlplane',
                                'tripleo_role=CustomRole',
                                'tripleo_stack=overcloud',
                                'tripleo_default_route=True'],
                          ),
]



#!/usr/bin/env python

# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import OrderedDict
import copy
import logging
import os
import sys
import tempfile
import yaml

from heatclient.exc import HTTPNotFound
import keystoneauth1
import openstack

from tripleo_common import exception
import tripleo_common.constants as constants

HOST_NETWORK = 'ctlplane'
DEFAULT_DOMAIN = 'localdomain.'

UNDERCLOUD_CONNECTION_SSH = 'ssh'

UNDERCLOUD_CONNECTION_LOCAL = 'local'

logging.basicConfig()
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class TemplateDumper(yaml.SafeDumper):
    def represent_ordered_dict(self, data):
        return self.represent_dict(data.items())


TemplateDumper.add_representer(OrderedDict,
                               TemplateDumper.represent_ordered_dict)


class StackOutputs(object):
    """Item getter for stack outputs.

    It takes a long time to resolve stack outputs.  This class ensures that
    we only have to do it once and then reuse the results from that call in
    subsequent lookups.  It also lazy loads the outputs so we don't spend time
    on unnecessary Heat calls.
    """

    def __init__(self, stack):
        self.outputs = {}
        self.stack = stack

    def _load_outputs(self):
        """Load outputs from the stack if necessary

        Retrieves the stack outputs if that has not already happened.  If it
        has then this is a noop.
        """
        if not self.outputs:
            self.outputs = {i['output_key']: i['output_value']
                            for i in self.stack.outputs
                            }

    def __getitem__(self, key):
        self._load_outputs()
        return self.outputs[key]

    def __iter__(self):
        self._load_outputs()
        return iter(self.outputs.keys())

    def get(self, key, default=None):
        try:
            self.__getitem__(key)
        except KeyError:
            pass
        return self.outputs.get(key, default)


class NeutronData(object):
    """Neutron inventory data.

    A data object with for inventory generation enriched neutron data.
    """
    def __init__(self, networks, subnets, ports, host_network=None):
        self.networks = networks
        self.subnets = subnets
        self.ports = ports
        self.host_network = host_network or HOST_NETWORK
        self.networks_by_id = self._networks_by_id()
        self.subnets_by_id = self._subnets_by_id()
        self.ports_by_role_and_host = self._ports_by_role_and_host()

    def _tags_to_dict(self, tags):
        tag_dict = dict()
        for tag in tags:
            if not tag.startswith('tripleo_'):
                continue
            try:
                key, value = tag.rsplit('=')
            except ValueError:
                continue

            # Make booleans type bool
            value = True if value in {'True', 'true', True} else value
            value = False if value in {'False', 'false', False} else value

            # Convert network index value to integer
            if key == 'tripleo_net_idx':
                value = int(value)

            tag_dict.update({key: value})

        return tag_dict

    def _ports_by_role_and_host(self):
        mandatory_tags = {'tripleo_role'}

        ports_by_role_and_host = {}
        for port in self.ports:
            tags = self._tags_to_dict(port.tags)
            # In case of missing required tags, raise an error.
            # neutron is useless as a inventory source in this case.
            if not mandatory_tags.issubset(tags):
                raise exception.MissingMandatoryNeutronResourceTag()

            hostname = port.dns_name
            network_id = port.network_id
            network = self.networks_by_id[network_id]
            fixed_ips = port.fixed_ips[0]
            subnet_id = fixed_ips.get('subnet_id')
            subnet = self.subnets_by_id[subnet_id]

            # "TripleO" cidr is the number of bits in the network mask
            cidr = subnet['cidr'].split('/')[1]
            dns_domain = network['dns_domain']
            dns_nameservers = subnet['dns_nameservers']
            mtu = network['mtu']
            net_name = network['name']
            ip_address = fixed_ips.get('ip_address')
            gateway_ip = subnet['gateway_ip']
            # Need deepcopy here so that adding default entry does not end
            # up in the subnet object and leak to other nodes with a different
            # default route network.
            host_routes = copy.deepcopy(subnet['host_routes'])
            # If this is the default route network, add a default route using
            # gateway_ip to the host_routes unless it's already present
            if tags.get('tripleo_default_route'):
                host_routes.append({'default': True, 'nexthop': gateway_ip})

            vlan_id = subnet['tags'].get('tripleo_vlan_id',
                                         constants.DEFAULT_VLAN_ID)
            role_name = tags['tripleo_role']

            role = ports_by_role_and_host.setdefault(role_name, {})
            host = role.setdefault(hostname, [])
            host.append(
                {'name': port.name,
                 'hostname': hostname,
                 'dns_domain': dns_domain,
                 'network_id': network_id,
                 'network_name': net_name,
                 'fixed_ips': port.fixed_ips,
                 'subnet_id': subnet_id,
                 'ip_address': ip_address,
                 'mtu': mtu,
                 'cidr': cidr,
                 'gateway_ip': gateway_ip,
                 'dns_nameservers': dns_nameservers,
                 'host_routes': host_routes,
                 'vlan_id': vlan_id,
                 'tags': tags}
            )

        return ports_by_role_and_host

    def _networks_by_id(self):
        mandatory_tags = {'tripleo_network_name'}
        networks_by_id = {}
        for net in self.networks:
            tags = self._tags_to_dict(net.tags)
            # In case of missing required tags, raise an error.
            # neutron is useless as a inventory source in this case.
            if (net.name != self.host_network and
                    not mandatory_tags.issubset(tags)):
                raise exception.MissingMandatoryNeutronResourceTag()

            if net.name != self.host_network:
                name_upper = tags['tripleo_network_name']
            else:
                name_upper = self.host_network
            networks_by_id.update(
                {net.id: {'name': net.name,
                          'name_upper': name_upper,
                          'subnet_ids': net.subnet_ids,
                          'mtu': net.mtu,
                          'dns_domain': net.dns_domain,
                          'tags': tags}
                 }
            )

        return networks_by_id

    def _subnets_by_id(self):
        subnets_by_id = {}
        for subnet in self.subnets:
            subnets_by_id.update(
                {subnet.id: {'name': subnet.name,
                             'network_id': subnet.network_id,
                             'ip_version': subnet.ip_version,
                             'gateway_ip': subnet.gateway_ip,
                             'cidr': subnet.cidr,
                             'host_routes': subnet.host_routes,
                             'dns_nameservers': subnet.dns_nameservers,
                             'tags': self._tags_to_dict(subnet.tags)}
                 }
            )

        return subnets_by_id


class TripleoInventory(object):
    def __init__(self, session=None, hclient=None,
                 plan_name=None, auth_url=None, project_name=None,
                 cacert=None, username=None, ansible_ssh_user=None,
                 host_network=None, ansible_python_interpreter=None,
                 undercloud_connection=UNDERCLOUD_CONNECTION_LOCAL,
                 undercloud_key_file=None, serial=1, work_dir=None):
        self.session = session
        self.hclient = hclient
        self.host_network = host_network or HOST_NETWORK
        self.auth_url = auth_url
        self.cacert = cacert
        self.project_name = project_name
        self.username = username
        self.ansible_ssh_user = ansible_ssh_user
        self.undercloud_key_file = undercloud_key_file
        self.plan_name = plan_name
        self.ansible_python_interpreter = ansible_python_interpreter
        self.hostvars = {}
        self.undercloud_connection = undercloud_connection
        self.serial = serial
        self.work_dir = work_dir

    @staticmethod
    def get_roles_by_service(enabled_services):
        # Flatten the lists of services for each role into a set
        services = set(
            [item for role_services in enabled_services.values()
             for item in role_services])

        roles_by_services = {}
        for service in services:
            roles_by_services[service] = []
            for role, val in enabled_services.items():
                if service in val:
                    roles_by_services[service].append(role)
            roles_by_services[service] = sorted(roles_by_services[service])
        return roles_by_services

    def get_overcloud_environment(self):
        try:
            environment = self.hclient.stacks.environment(self.plan_name)
            return environment
        except (HTTPNotFound,
                keystoneauth1.exceptions.catalog.EndpointNotFound):
            return {}

    UNDERCLOUD_SERVICES = [
        'tripleo_nova_compute', 'tripleo_heat_engine',
        'tripleo_ironic_conductor', 'tripleo_swift_container_server',
        'tripleo_swift_object_server', 'tripleo_mistral_engine']

    def get_undercloud_service_list(self):
        """Return list of undercloud services - currently static

        Replace this when we have a better way - e.g. heat deploys undercloud
        """
        return self.UNDERCLOUD_SERVICES

    def _hosts(self, alist, dynamic=True):
        """Static yaml inventories reqire a different hosts format?!"""
        if not dynamic:
            return {x: {} for x in alist}
        return alist

    def _get_stack(self):
        if self.plan_name is None:
            return None
        try:
            stack = self.hclient.stacks.get(self.plan_name)
        except (HTTPNotFound,
                keystoneauth1.exceptions.catalog.EndpointNotFound):
            stack = None

        return stack

    def _inventory_from_heat_outputs(self, ret, children, dynamic):
        if not self.stack:
            return

        vip_map = self.stack_outputs.get('VipMap', {})
        role_net_ip_map = self.stack_outputs.get('RoleNetIpMap', {})
        role_node_id_map = self.stack_outputs.get('ServerIdData', {})
        networks = set()
        role_net_hostname_map = self.stack_outputs.get(
            'RoleNetHostnameMap', {})
        for role_name, hostnames in role_net_hostname_map.items():
            if not hostnames:
                continue

            net_ip_map = role_net_ip_map[role_name]
            ips = net_ip_map[self.host_network]
            if not ips:
                raise Exception("No IPs found for %s role on %s network" %
                                (role_name, self.host_network))

            net_hostname_map = role_net_hostname_map[role_name]
            bootstrap_server_id = role_node_id_map.get('bootstrap_server_id')
            node_id_map = role_node_id_map.get('server_ids')
            if node_id_map:
                srv_id_map = node_id_map.get(role_name)

            role_networks = sorted([str(net) for net in net_ip_map])
            networks.update(role_networks)

            # Undercloud role in the stack should overwrite, not append.
            # See bug: https://bugs.launchpad.net/tripleo/+bug/1913551
            if role_name == 'Undercloud':
                role = ret[role_name] = {}
            else:
                role = ret.setdefault(role_name, {})

            hosts = role.setdefault('hosts', {})
            role_vars = role.setdefault('vars', {})

            role_vars.setdefault('ansible_ssh_user', self.ansible_ssh_user)
            role_vars.setdefault('bootstrap_server_id', bootstrap_server_id)
            role_vars.setdefault('tripleo_role_name', role_name)
            role_vars.setdefault('tripleo_role_networks', role_networks)
            role_vars.setdefault('serial', self.serial)

            if self.ansible_python_interpreter:
                role_vars.setdefault('ansible_python_interpreter',
                                     self.ansible_python_interpreter)

            names = hostnames.get(self.host_network) or []
            shortnames = [n.split(".%s." % self.host_network)[0].lower()
                          for n in names]

            for idx, name in enumerate(shortnames):
                host = hosts.setdefault(name, {})
                host.setdefault('ansible_host', ips[idx])

                if srv_id_map:
                    host.setdefault('deploy_server_id', srv_id_map[idx])

                # Add variable for IP on each network
                for net in net_ip_map:
                    host.setdefault('{}_ip'.format(net), net_ip_map[net][idx])

                # Add variables for hostname on each network
                for net in net_hostname_map:
                    host.setdefault(
                        '{}_hostname'.format(net), net_hostname_map[net][idx])

            children.add(role_name)

            self.hostvars.update(hosts)

            if dynamic:
                hosts_format = [h for h in hosts.keys()]
                hosts_format.sort()
                ret[role_name]['hosts'] = hosts_format

        if children:
            allovercloud = ret.setdefault('allovercloud', {})
            overcloud_vars = allovercloud.setdefault('vars', {})

            for vip_name, vip in vip_map.items():
                if vip and (vip_name in networks or vip_name == 'redis'):
                    overcloud_vars.setdefault('{}_vip'.format(vip_name), vip)

            overcloud_vars.setdefault(
                'container_cli', self.get_overcloud_environment().get(
                    'parameter_defaults', {}).get('ContainerCli'))

            allovercloud.setdefault('children', self._hosts(sorted(children),
                                                            dynamic))

            ret.setdefault(
                self.plan_name, {'children': self._hosts(['allovercloud'],
                                                         dynamic)})

            if self.plan_name != 'overcloud':
                ret.setdefault('overcloud',
                               {'children': self._hosts(['allovercloud'],
                                                        dynamic),
                                'deprecated': ('Deprecated by allovercloud '
                                               'group in Ussuri')})

        # Associate services with roles
        roles_by_service = self.get_roles_by_service(
            self.stack_outputs.get('EnabledServices', {}))

        # tripleo-groups map to ceph-ansible groups as follows
        ceph_group_map = {
            'ceph_mon': 'mons',
            'ceph_osd': 'osds',
            'ceph_mgr': 'mgrs',
            'ceph_rgw': 'rgws',
            'ceph_mds': 'mdss',
            'ceph_nfs': 'nfss',
            'ceph_client': 'clients',
            'ceph_rbdmirror': 'rbdmirrors',
            'ceph_grafana': 'grafana-server'
        }
        # add a ceph-ansible compatible group to the inventory
        # which has the same roles. E.g. if the inventory has
        # a group 'ceph_mon' which has childen and vars, then
        # the inventory will now also have a group 'mons' with
        # the same children and vars.
        for service, roles in roles_by_service.copy().items():
            if service in ceph_group_map.keys():
                roles_by_service[ceph_group_map[service]] = roles

        for service, roles in roles_by_service.items():
            service_children = [role for role in roles
                                if ret.get(role) is not None]
            if service_children:
                svc_host = ret.setdefault(service.lower(), {})
                svc_host_vars = svc_host.setdefault('vars', {})
                svc_host.setdefault('children', self._hosts(service_children,
                                                            dynamic))
                svc_host_vars.setdefault('ansible_ssh_user',
                                         self.ansible_ssh_user)
                if self.ansible_python_interpreter:
                    svc_host_vars.setdefault('ansible_python_interpreter',
                                             self.ansible_python_interpreter)

    def _get_neutron_data(self):
        if not self.session:
            LOG.info("Session not set, neutron data will not be used to build "
                     "the inventory.")
            return

        try:
            conn = openstack.connection.Connection(session=self.session)
            tags_filter = ['tripleo_stack_name={}'.format(self.plan_name)]
            ports = list(conn.network.ports(tags=tags_filter))
            if not ports:
                return None

            networks = [conn.network.find_network(p.network_id)
                        for p in ports]
            subnets = []
            for net in networks:
                subnets.extend(conn.network.subnets(network_id=net.id))

            data = NeutronData(networks, subnets, ports)
        except exception.MissingMandatoryNeutronResourceTag:
            # In case of missing required tags, neutron is useless as an
            # inventory source, log warning and return None to disable the
            # neutron source.
            LOG.warning("Neutron resource without mandatory tags present. "
                        "Disabling use of neutron as a source for inventory "
                        "generation.")
            return None
        except (openstack.connection.exceptions.EndpointNotFound,
                keystoneauth1.exceptions.catalog.EndpointNotFound):
            LOG.warning("Neutron service not installed. Disabling use of "
                        "neutron as a source for inventory generation.")
            return None

        return data

    def _add_host_from_neutron_data(self, host, ports, role_networks,
                                    role_vars):
        for port in ports:
            net_name = port['network_name']

            # Add network name to tripleo_role_networks variable
            if net_name not in role_networks:
                role_networks.append(net_name)

            # Append to role_vars if not already present
            net_config_keys = {'cidr', 'dns_nameservers', 'gateway_ip',
                               'host_routes', 'vlan_id'}
            for key in net_config_keys:
                var = '{}_{}'.format(net_name, key)
                if var not in role_vars:
                    role_vars.setdefault(var, port[key])

            # Add variable for hostname on network
            host.setdefault('{}_hostname'.format(net_name), '.'.join(
                [port['hostname'], port['dns_domain']]))

            # Add variable for IP address on networks
            host.setdefault('{}_ip'.format(net_name), port['ip_address'])

            if net_name == self.host_network:
                # Add variable for ansible_host
                host.setdefault('ansible_host', port['ip_address'])

                # Add variable for canonical hostname
                dns_domain = port.get('dns_domain')
                if dns_domain:
                    canonical_dns_domain = dns_domain.partition('.')[-1]
                else:
                    canonical_dns_domain = DEFAULT_DOMAIN
                host.setdefault('canonical_hostname', '.'.join(
                    [port['hostname'], canonical_dns_domain]))

    def _inventory_from_neutron_data(self, ret, children, dynamic):
        if not self.neutron_data:
            return
        ports_by_role_and_host = self.neutron_data.ports_by_role_and_host
        networks_by_id = self.neutron_data.networks_by_id

        netname_by_idx = {
            net['tags'].get('tripleo_net_idx'):
                net['tags'].get('tripleo_network_name')
            for _, net in networks_by_id.items()
            if net['name'] != self.host_network}
        networks_all = [netname_by_idx[idx] for idx in sorted(netname_by_idx)]
        networks_lower = {net['name_upper']: net['name']
                          for _, net in networks_by_id.items()}
        networks_upper = {net['name']: net['name_upper']
                          for _, net in networks_by_id.items()}

        for role_name, ports_by_host in ports_by_role_and_host.items():
            role = ret.setdefault(role_name, {})
            hosts = role.setdefault('hosts', {})
            role_vars = role.setdefault('vars', {})
            role_vars.setdefault('tripleo_role_name', role_name)
            role_vars.setdefault('ansible_ssh_user', self.ansible_ssh_user)
            role_vars.setdefault('serial', self.serial)
            role_networks = role_vars.setdefault('tripleo_role_networks', [])
            for hostname, ports in ports_by_host.items():
                host = hosts.setdefault(hostname, {})
                self._add_host_from_neutron_data(host, ports, role_networks,
                                                 role_vars)

            # The nic config templates use ctlplane_subnet_cidr, not
            # ctlplane_cidr. Handle the special case.
            role_vars.setdefault(self.host_network + '_subnet_cidr',
                                 role_vars[self.host_network + '_cidr'])
            role_vars.setdefault('tripleo_role_networks',
                                 sorted(role_networks))
            role_vars.setdefault(
                'role_networks',
                [networks_upper[net] for net in role_networks])
            role_vars.setdefault('networks_all', networks_all)
            role_vars.setdefault('networks_lower', networks_lower)

            for _, net in networks_by_id.items():
                role_vars.setdefault(net['name'] + '_mtu', net['mtu'])

            children.add(role_name)
            self.hostvars.update(hosts)

            if dynamic:
                hosts_format = [h for h in hosts.keys()]
                hosts_format.sort()
                ret[role_name]['hosts'] = hosts_format

        if children:
            allovercloud = ret.setdefault('allovercloud', {})
            allovercloud.setdefault('children',
                                    self._hosts(sorted(children), dynamic))

    def _extend_inventory(self, ret, dynamic, data=None):
        if not data:
            return

        for role_name, role_values in data.items():
            inventory_role = ret.get(role_name)
            if not inventory_role:
                continue
            inventory_hosts = inventory_role.get('hosts', {})
            inventory_vars = inventory_role.get('vars', {})

            config_file_hosts = role_values.get('hosts', {})
            config_file_vars = role_values.get('vars', {})

            for k, v in config_file_vars.items():
                inventory_vars.setdefault(k, v)

            for config_file_host, host_values in config_file_hosts.items():
                inventory_host = inventory_hosts.get(config_file_host, {})
                if not inventory_host:
                    continue

                for k, v in host_values.items():
                    inventory_host.setdefault(k, v)

            self.hostvars.update(inventory_hosts)
            if dynamic:
                hosts_format = [h for h in inventory_hosts.keys()]
                hosts_format.sort()
                ret[role_name]['hosts'] = hosts_format

    def _get_data_from_config_file(self):
        if not self.plan_name:
            return
        if not self.work_dir:
            return

        data_file_path = os.path.join(self.work_dir,
                                      constants.INVENTORY_NETWORK_CONFIG_FILE)
        if not os.path.isfile(data_file_path):
            return
        with open(data_file_path, 'r') as f:
            data = yaml.safe_load(f.read())

        return data

    def _undercloud_inventory(self, ret, dynamic):
        undercloud = ret.setdefault('Undercloud', {})
        undercloud.setdefault('hosts', self._hosts(['undercloud'], dynamic))
        _vars = undercloud.setdefault('vars', {})
        _vars.setdefault('ansible_host', 'localhost')
        _vars.setdefault('ansible_connection', self.undercloud_connection)
        # see https://github.com/ansible/ansible/issues/41808
        _vars.setdefault('ansible_remote_tmp', '/tmp/ansible-${USER}')
        _vars.setdefault('auth_url', self.auth_url)
        _vars.setdefault('project_name', self.project_name)
        _vars.setdefault('username', self.username)

        if self.cacert:
            _vars['cacert'] = self.cacert

        if self.ansible_python_interpreter:
            _vars.setdefault('ansible_python_interpreter',
                             self.ansible_python_interpreter)
        else:
            _vars.setdefault('ansible_python_interpreter', sys.executable)

        if self.undercloud_connection == UNDERCLOUD_CONNECTION_SSH:
            _vars.setdefault('ansible_ssh_user', self.ansible_ssh_user)
            if self.undercloud_key_file:
                _vars.setdefault('ansible_ssh_private_key_file',
                                 self.undercloud_key_file)

        _vars.setdefault('undercloud_service_list',
                         self.get_undercloud_service_list())

        # Remaining variables need the stack to be resolved ...
        if not self.stack:
            return

        _vars.setdefault('plan', self.plan_name)

        admin_password = self.get_overcloud_environment().get(
            'parameter_defaults', {}).get('AdminPassword')
        if admin_password:
            _vars.setdefault('overcloud_admin_password', admin_password)

        keystone_url = self.stack_outputs.get('KeystoneURL')
        if keystone_url:
            _vars.setdefault('overcloud_keystone_url', keystone_url)

        endpoint_map = self.stack_outputs.get('EndpointMap')
        if endpoint_map:
            horizon_endpoint = endpoint_map.get('HorizonPublic', {}).get('uri')
            if horizon_endpoint:
                _vars.setdefault('overcloud_horizon_url', horizon_endpoint)

    def list(self, dynamic=True):
        ret = OrderedDict()
        if dynamic:
            # Prevent Ansible from repeatedly calling us to get empty host
            # details
            ret.setdefault('_meta', {'hostvars': self.hostvars})

        children = set()

        self.stack = self._get_stack()
        self.stack_outputs = StackOutputs(self.stack)

        self.neutron_data = self._get_neutron_data()

        if self.stack is None and self.neutron_data is None:
            LOG.warning("Stack not found: %s. No data found in neither "
                        "neutron or heat. Only the undercloud will be added "
                        "to the inventory.", self.plan_name)

        self._undercloud_inventory(ret, dynamic)
        self._inventory_from_neutron_data(ret, children, dynamic)
        self._inventory_from_heat_outputs(ret, children, dynamic)
        self._extend_inventory(ret, dynamic,
                               data=self._get_data_from_config_file())

        return ret

    def host(self):
        # NOTE(mandre)
        # Dynamic inventory scripts must return empty json if they don't
        # provide detailed info for hosts:
        # http://docs.ansible.com/ansible/developing_inventory.html
        return {}

    def write_static_inventory(self, inventory_file_path, extra_vars=None):
        """Convert inventory list to static yaml format in a file."""
        allowed_extensions = ('.yaml', '.yml', '.json')
        if not os.path.splitext(inventory_file_path)[1] in allowed_extensions:
            raise ValueError("Path %s does not end with one of %s extensions"
                             % (inventory_file_path,
                                ",".join(allowed_extensions)))

        # For some reason the json/yaml format needed for static and
        # dynamic inventories is different for the hosts/children?!
        inventory = self.list(dynamic=False)

        if extra_vars:
            for var, value in extra_vars.items():
                if var in inventory:
                    inventory[var]['vars'].update(value)

        # Atomic update as concurrent tripleoclient commands can call this
        inventory_file_dir = os.path.dirname(inventory_file_path)
        with tempfile.NamedTemporaryFile(
                'w',
                dir=inventory_file_dir,
                delete=False) as inventory_file:
            yaml.dump(inventory, inventory_file, TemplateDumper)
        os.rename(inventory_file.name, inventory_file_path)


def generate_tripleo_ansible_inventory(heat, auth_url,
                                       username,
                                       project_name,
                                       cacert=None,
                                       plan='overcloud',
                                       work_dir=None,
                                       ansible_python_interpreter=None,
                                       ansible_ssh_user='tripleo-admin',
                                       undercloud_key_file=None,
                                       ssh_network='ctlplane',
                                       session=None):
    if not work_dir:
        work_dir = os.path.join(os.path.expanduser('~'),
                                'overcloud-deploy-{}'.format(plan))
    if not os.path.isdir(work_dir):
        work_dir = tempfile.mkdtemp(prefix='tripleo-ansible')

    inventory_path = os.path.join(
        work_dir, 'tripleo-ansible-inventory.yaml')
    inv = TripleoInventory(
        session=session,
        hclient=heat,
        auth_url=auth_url,
        username=username,
        project_name=project_name,
        cacert=cacert,
        ansible_ssh_user=ansible_ssh_user,
        undercloud_key_file=undercloud_key_file,
        ansible_python_interpreter=ansible_python_interpreter,
        plan_name=plan,
        host_network=ssh_network,
        work_dir=work_dir)

    inv.write_static_inventory(inventory_path)
    return inventory_path



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tripleo_common.utils import ansible as ansible_utils


# FIXME(ramishra) Left behind till we've a new tripleo-common release
def write_default_ansible_cfg(work_dir,
                              remote_user,
                              ssh_private_key=None,
                              transport=None,
                              base_ansible_cfg='/etc/ansible/ansible.cfg',
                              override_ansible_cfg=None):

    return ansible_utils.write_default_ansible_cfg(
        work_dir, remote_user,
        ssh_private_key, transport,
        base_ansible_cfg,
        override_ansible_cfg)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from datetime import datetime
import json
import logging
import multiprocessing
import os
from pathlib import Path
import shutil
import six
from six.moves import configparser
from six.moves import cStringIO as StringIO
import sys
import tempfile
import yaml

from oslo_concurrency import processutils

from tripleo_common import constants

LOG = logging.getLogger(__name__)


def write_default_ansible_cfg(work_dir,
                              remote_user,
                              ssh_private_key=None,
                              transport=None,
                              base_ansible_cfg='/etc/ansible/ansible.cfg',
                              override_ansible_cfg=None):
    ansible_config_path = os.path.join(work_dir, 'ansible.cfg')
    shutil.copy(base_ansible_cfg, ansible_config_path)

    modules_path = (
        '/root/.ansible/plugins/modules:'
        '/usr/share/ansible/tripleo-plugins/modules:'
        '/usr/share/ansible/plugins/modules:'
        '/usr/share/ansible-modules:'
        '{}/library:'
        '{}/library'.format(constants.DEFAULT_VALIDATIONS_BASEDIR,
                            constants.DEFAULT_VALIDATIONS_LEGACY_BASEDIR))
    lookups_path = (
        '/root/.ansible/plugins/lookup:'
        '/usr/share/ansible/tripleo-plugins/lookup:'
        '/usr/share/ansible/plugins/lookup:'
        '{}/lookup_plugins:'
        '{}/lookup_plugins'.format(
            constants.DEFAULT_VALIDATIONS_BASEDIR,
            constants.DEFAULT_VALIDATIONS_LEGACY_BASEDIR))
    callbacks_path = (
        '~/.ansible/plugins/callback:'
        '/usr/share/ansible/tripleo-plugins/callback:'
        '/usr/share/ansible/plugins/callback:'
        '{}/callback_plugins:'
        '{}/callback_plugins'.format(
            constants.DEFAULT_VALIDATIONS_BASEDIR,
            constants.DEFAULT_VALIDATIONS_LEGACY_BASEDIR))

    callbacks_whitelist = ','.join(['tripleo_dense', 'tripleo_profile_tasks',
                                    'tripleo_states'])
    action_plugins_path = (
        '~/.ansible/plugins/action:'
        '/usr/share/ansible/plugins/action:'
        '/usr/share/ansible/tripleo-plugins/action:'
        '{}/action_plugins:'
        '{}/action_plugins'.format(
            constants.DEFAULT_VALIDATIONS_BASEDIR,
            constants.DEFAULT_VALIDATIONS_LEGACY_BASEDIR))
    filter_plugins_path = (
        '~/.ansible/plugins/filter:'
        '/usr/share/ansible/plugins/filter:'
        '/usr/share/ansible/tripleo-plugins/filter:'
        '{}/filter_plugins:'
        '{}/filter_plugins'.format(
            constants.DEFAULT_VALIDATIONS_BASEDIR,
            constants.DEFAULT_VALIDATIONS_LEGACY_BASEDIR))
    roles_path = ('%(work_dir)s/roles:'
                  '/root/.ansible/roles:'
                  '/usr/share/ansible/tripleo-roles:'
                  '/usr/share/ansible/roles:'
                  '/etc/ansible/roles:'
                  '%(ooo_val_path)s/roles:'
                  '%(work_dir)s' % {
                      'work_dir': work_dir,
                      'ooo_val_path':
                          constants.DEFAULT_VALIDATIONS_LEGACY_BASEDIR
                  })

    config = configparser.ConfigParser()
    config.read(ansible_config_path)

    config.set('defaults', 'retry_files_enabled', 'False')
    config.set('defaults', 'roles_path', roles_path)
    config.set('defaults', 'library', modules_path)
    config.set('defaults', 'callback_plugins', callbacks_path)
    config.set('defaults', 'callback_whitelist', callbacks_whitelist)
    config.set('defaults', 'stdout_callback', 'tripleo_dense')
    config.set('defaults', 'action_plugins', action_plugins_path)
    config.set('defaults', 'lookup_plugins', lookups_path)
    config.set('defaults', 'filter_plugins', filter_plugins_path)

    log_path = os.path.join(work_dir, 'ansible.log')
    config.set('defaults', 'log_path', log_path)
    if os.path.exists(log_path):
        new_path = (log_path + '-' +
                    datetime.now().strftime("%Y-%m-%dT%H:%M:%S"))
        os.rename(log_path, new_path)

    # Create the log file, and set some rights on it in order to prevent
    # unwanted accesse
    Path(log_path).touch()
    os.chmod(log_path, 0o640)

    config.set('defaults', 'forks', str(min(
        multiprocessing.cpu_count() * 4, 100)))
    config.set('defaults', 'timeout', '30')
    config.set('defaults', 'gather_timeout', '30')

    # Setup fact cache to improve playbook execution speed
    config.set('defaults', 'gathering', 'smart')
    config.set('defaults', 'fact_caching', 'jsonfile')
    config.set('defaults', 'fact_caching_connection',
               '~/.ansible/fact_cache')
    # NOTE(mwhahaha): only gather the bare minimum facts because this has
    # direct impact on how fast ansible can go.
    config.set('defaults', 'gather_subset', '!all,min')
    # NOTE(mwhahaha): this significantly affects performation per ansible#73654
    config.set('defaults', 'inject_facts_as_vars', 'false')

    # Set the pull interval to lower CPU overhead
    config.set('defaults', 'internal_poll_interval', '0.01')

    # Set the interpreter discovery to auto mode.
    config.set('defaults', 'interpreter_python', 'auto')

    # Expire facts in the fact cache after 7200s (2h)
    config.set('defaults', 'fact_caching_timeout', '7200')

    # mistral user has no home dir set, so no place to save a known hosts file
    config.set('ssh_connection', 'ssh_args',
               '-o UserKnownHostsFile=/dev/null '
               '-o StrictHostKeyChecking=no '
               '-o ControlMaster=auto '
               '-o ControlPersist=30m '
               '-o ServerAliveInterval=5 '
               '-o ServerAliveCountMax=5 '
               '-o PreferredAuthentications=publickey')
    config.set('ssh_connection', 'control_path_dir',
               os.path.join(work_dir, 'ansible-ssh'))
    config.set('ssh_connection', 'retries', '8')
    config.set('ssh_connection', 'pipelining', 'True')
    # Related to https://github.com/ansible/ansible/issues/22127
    config.set('ssh_connection', 'scp_if_ssh', 'True')

    if override_ansible_cfg:
        sio_cfg = StringIO()
        sio_cfg.write(override_ansible_cfg)
        sio_cfg.seek(0)
        config.read_file(sio_cfg)
        sio_cfg.close()

    with open(ansible_config_path, 'w') as configfile:
        config.write(configfile)

    return ansible_config_path


def _get_inventory(inventory, work_dir):
    if not inventory:
        return None

    if (isinstance(inventory, six.string_types) and
            os.path.exists(inventory)):
        return inventory
    if not isinstance(inventory, six.string_types):
        inventory = yaml.safe_dump(inventory)

    path = os.path.join(work_dir, 'inventory.yaml')

    with open(path, 'w') as inv:
        inv.write(inventory)

    return path


def _get_ssh_private_key(ssh_private_key, work_dir):
    if not ssh_private_key:
        return None

    if (isinstance(ssh_private_key, six.string_types) and
            os.path.exists(ssh_private_key)):
        os.chmod(ssh_private_key, 0o600)
        return ssh_private_key

    path = os.path.join(work_dir, 'ssh_private_key')

    with open(path, 'w') as ssh_key:
        ssh_key.write(ssh_private_key)
    os.chmod(path, 0o600)

    return path


def _get_playbook(playbook, work_dir):
    if not playbook:
        return None

    if (isinstance(playbook, six.string_types) and
            os.path.exists(playbook)):
        return playbook
    if not isinstance(playbook, six.string_types):
        playbook = yaml.safe_dump(playbook)

    path = os.path.join(work_dir, 'playbook.yaml')

    with open(path, 'w') as pb:
        pb.write(playbook)

    return path


def run_ansible_playbook(playbook, work_dir=None, **kwargs):
    verbosity = kwargs.get('verbosity', 5)
    remove_work_dir = False
    if not work_dir:
        work_dir = tempfile.mkdtemp(prefix='tripleo-ansible')
        remove_work_dir = True

    playbook = _get_playbook(playbook, work_dir)
    python_version = sys.version_info.major
    ansible_playbook_cmd = "ansible-playbook-{}".format(python_version)
    if 1 < verbosity < 6:
        verbosity_option = '-' + ('v' * (verbosity - 1))
        command = [ansible_playbook_cmd, verbosity_option,
                   playbook]
    else:
        command = [ansible_playbook_cmd, playbook]

    limit_hosts = kwargs.get('limit_hosts', None)
    if limit_hosts:
        command.extend(['--limit', limit_hosts])

    module_path = kwargs.get('module_path', None)
    if module_path:
        command.extend(['--module-path', module_path])

    become = kwargs.get('become', False)
    if become:
        command.extend(['--become'])

    become_user = kwargs.get('become_user', None)
    if become_user:
        command.extend(['--become-user', become_user])

    extra_vars = kwargs.get('extra_vars', None)
    if extra_vars:
        extra_vars = json.dumps(extra_vars)
        command.extend(['--extra-vars', extra_vars])

    flush_cache = kwargs.get('flush_cache', False)
    if flush_cache:
        command.extend(['--flush-cache'])

    forks = kwargs.get('forks', None)
    if forks:
        command.extend(['--forks', forks])

    ssh_common_args = kwargs.get('ssh_common_args', None)
    if ssh_common_args:
        command.extend(['--ssh-common-args', ssh_common_args])

    ssh_extra_args = kwargs.get('ssh_extra_args', None)
    if ssh_extra_args:
        command.extend(['--ssh-extra-args', ssh_extra_args])

    timeout = kwargs.get('timeout', None)
    if timeout:
        command.extend(['--timeout', timeout])

    inventory = _get_inventory(kwargs.get('inventory', None),
                               work_dir)
    if inventory:
        command.extend(['--inventory-file', inventory])

    tags = kwargs.get('tags', None)
    if tags:
        command.extend(['--tags', tags])

    skip_tags = kwargs.get('skip_tags', None)
    if skip_tags:
        command.extend(['--skip-tags', skip_tags])

    extra_env_variables = kwargs.get('extra_env_variables', None)
    override_ansible_cfg = kwargs.get('override_ansible_cfg', None)
    remote_user = kwargs.get('remote_user', None)
    ssh_private_key = kwargs.get('ssh_private_key', None)

    if extra_env_variables:
        if not isinstance(extra_env_variables, dict):
            msg = "extra_env_variables must be a dict"
            raise RuntimeError(msg)
        for key, value in extra_env_variables.items():
            extra_env_variables[key] = six.text_type(value)

    try:
        ansible_config_path = write_default_ansible_cfg(
            work_dir,
            remote_user,
            ssh_private_key=_get_ssh_private_key(
                ssh_private_key, work_dir),
            override_ansible_cfg=override_ansible_cfg)
        env_variables = {
            'HOME': work_dir,
            'ANSIBLE_LOCAL_TEMP': work_dir,
            'ANSIBLE_CONFIG': ansible_config_path,
        }

        profile_tasks = kwargs.get('profile_tasks', True)
        if profile_tasks:
            profile_tasks_limit = kwargs.get('profile_tasks_limit', 20)
            env_variables.update({
                # the whitelist could be collected from multiple
                # arguments if we find a use case for it
                'ANSIBLE_CALLBACK_WHITELIST':
                    'tripleo_dense,tripleo_profile_tasks,tripleo_states',
                'ANSIBLE_STDOUT_CALLBACK': 'tripleo_dense',
                'PROFILE_TASKS_TASK_OUTPUT_LIMIT':
                    six.text_type(profile_tasks_limit),
            })

        if extra_env_variables:
            env_variables.update(extra_env_variables)

        command = [str(c) for c in command]

        reproduce_command = kwargs.get('reproduce_command', None)
        command_timeout = kwargs.get('command_timeout', None)
        trash_output = kwargs.get('trash_output', None)
        if reproduce_command:
            command_path = os.path.join(work_dir,
                                        "ansible-playbook-command.sh")
            with open(command_path, 'w') as f:
                f.write('#!/bin/bash\n')
                f.write('\n')
                for var in env_variables:
                    f.write('%s="%s"\n' % (var, env_variables[var]))
                f.write('\n')
                f.write(' '.join(command))
                f.write(' "$@"')
                f.write('\n')

            os.chmod(command_path, 0o750)

        if command_timeout:
            command = ['timeout', '-s', 'KILL',
                       str(command_timeout)] + command

        LOG.info('Running ansible-playbook command: %s', command)

        stderr, stdout = processutils.execute(
            *command, cwd=work_dir,
            env_variables=env_variables,
            log_errors=processutils.LogErrors.ALL)
        if trash_output:
            stdout = ""
            stderr = ""
        return {"stderr": stderr, "stdout": stdout,
                "log_path": os.path.join(work_dir, 'ansible.log')}
    finally:
        try:
            if remove_work_dir:
                shutil.rmtree(work_dir)
        except Exception as e:
            msg = "An error happened while cleaning work directory: " + e
            raise RuntimeError(msg)
